#!/usr/bin/perl
my $buildno = '3.0.2008.09.27';
#
# Kakusan3
# This is a script to select nucleotide substitution model for multi-
# partitioned data set. Official web site of this script is
# http://www.fifthdimension.jp/products/kakusan/ .
# To know script details, see above URL.
# 
# Copyright (C) 2006-2008  Akifumi S. Tanabe
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

use strict;

$| = 1;

my $readseq = 'readseq';
my $paupcmd = 'paup';
my $baseml = 'baseml';
my $tf = 'tf';
my $dnadist = 'dnadist';
my $neighbor = 'neighbor';
my $dnapars = 'dnapars';

my $lnlcalc = 'baseml';

my @allmodel = ('JC69', 'F81', 'K80', 'HKY85', 'TN93ef', 'TN93', 'K81', 'K81uf', 'J1ef', 'J1', 'J2ef', 'J2', 'TIMef', 'TIM', 'TVMef', 'TVM', 'SYM', 'GTR');
my @allheterogeneity = ('Homogeneous', 'Gamma', 'Invariant', 'GammaInvariant', 'AGamma', 'CodonPos', 'CodonPos1Gamma', 'CodonPos3Gamma', 'CodonPos1AGamma', 'CodonPos3AGamma', 'Gene', 'Gene1Gamma', 'GeneNGamma', 'Gene1AGamma', 'GeneNAGamma', 'GeneCodonPos', 'GeneCodonPos1Gamma', 'GeneCodonPosNGamma', 'GeneCodonPos1AGamma', 'GeneCodonPosNAGamma');

my @modelformrbayes = ('JC69', 'F81', 'K80', 'HKY85', 'SYM', 'GTR');
my @heterogeneityformrbayes = ('Homogeneous', 'Gamma', 'Invariant', 'GammaInvariant', 'AGamma', 'CodonPos', 'CodonPos1Gamma', 'CodonPos3Gamma', 'CodonPos1AGamma', 'CodonPos3AGamma', 'Gene', 'Gene1Gamma', 'GeneNGamma', 'Gene1AGamma', 'GeneNAGamma', 'GeneCodonPos', 'GeneCodonPos1Gamma', 'GeneCodonPosNGamma', 'GeneCodonPos1AGamma', 'GeneCodonPosNAGamma');

my @modelforpaup = ('JC69', 'F81', 'K80', 'HKY85', 'TN93ef', 'TN93', 'K81', 'K81uf', 'J1ef', 'J1', 'J2ef', 'J2', 'TIMef', 'TIM', 'TVMef', 'TVM', 'SYM', 'GTR');
my @heterogeneityforpaup = ('Homogeneous', 'Gamma', 'Invariant', 'GammaInvariant', 'CodonPos', 'Gene', 'GeneCodonPos');

my @modelfortf = ('JC69', 'F81', 'K80', 'HKY85', 'TN93ef', 'TN93', 'J1ef', 'J1', 'J2ef', 'J2', 'TIMef', 'TIM', 'TVMef', 'TVM', 'SYM', 'GTR');
my @heterogeneityfortf = ('Homogeneous', 'Gamma', 'Invariant', 'GammaInvariant');

my $treefile;
my $interactive = 0;
my $chisq = 1;
my $siglevel = 0.05;
my $basefreq = 'Estimate';
my $ngammacat = 8;
my $n_gam = 0;
my $agamma = 0;
my $numthreads = 1;
my $difftreeeachpart;
my $algorithm = 'NJJC';
my $dividecodon = 0;
my $codonseparateonly = 0;
my $calcwhole = 0;
my $ntax;
my $nchar;
my $maindir;
my $taxlength = 6;
my @fullpathname;
my @inputfile;
my @taxa;
my @partname;
my @subpart;
my @allpart;
my %part2site;
my %subpart;
my %models;
my %heterogeneities;
my %model;
my %heterogeneity;
my %seqs;
my %nparam;
my %samplesize;
my %LnL;
my %criteria;
my %weight;
my %output;

# Main routine

print <<"_END";
Kakusan3 $buildno
=======================================================================

This is a script to select nucleotide substitution model for multi-
partitioned data set. Official web site of this script is
http://www.fifthdimension.jp/products/kakusan/ .
To know script details, see above URL.

Copyright (C) 2006-2008  Akifumi S. Tanabe

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

_END

# parse command line options
&parseOptions();

# check required and obstructive files
&checkFiles();

# convert sequences to non-interleaved NEXUS format
&convertSequences();

# split sequences to specified partitions
if (@inputfile == 1) {
	&splitSequences();
}
# combine sequences
else {
	&combineSequences();
}

# split protein coding sequences to each codon position
if (@partname != 0) {
	&splitCodingSequences();
}

# make each codon position files
if ($dividecodon && @subpart != 0) {
	&makeEachCodonPositions;
}

# set partitions
if (@partname) {
	&setPartitions();
}

# set candidate models
&setCandidateModels();

# run pre-selection chi-square test
&runChiSquareTest();

# set the number of model parameters
&setModelParameters();

# make starting tree
&makeStartingTree();

# set sample sizes
&setSampleSizes();

# analyze partitions
&analyzeEachPartition();

# calculate model selection criteria
&calculateCriteria();

# output files
&outputFiles();

# exit
if ($interactive) {
	print("The model selection have been finished.\nThe results was output to \"$maindir\".\nPlease press enter to close.\n");
	<STDIN>;
}
exit;

sub parseOptions {
	print("Parsing command line options...\n");
	my %infile;
	my @infile;
	my %site2part;
	foreach (@ARGV) {
		# -h option
		if (/^-+(?:help|h)/i) {
			&helpMessage();
		}
		# -i option
		elsif (/^-+(?:interactive|i)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$interactive = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$interactive = 0;
			}
			else {
				&errorMessage(__LINE__, 'Interactive mode option is not valid.');
			}
		}
		# -c option
		elsif (/^-+(?:chisq|c)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$chisq = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$chisq = 0;
			}
			else {
				&errorMessage(__LINE__, 'Chi-Square test option is not valid.');
			}
		}
		# -s option
		elsif (/^-+(?:siglevel|s)=(.+)/i) {
			if ($1 =~ /^(0\.?\d+)$/) {
				$siglevel = $1;
			}
			else {
				&errorMessage(__LINE__, 'The number of significant level is not valid.');
			}
			if ($siglevel == 0 || $siglevel >= 1) {
				&errorMessage(__LINE__, 'The number of significant level is not valid.');
			}
		}
		# -d option
		elsif (/^-+(?:dividecodon|d)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$dividecodon = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$dividecodon = 0;
			}
			else {
				&errorMessage(__LINE__, 'Codon position dividing option is not valid.');
			}
		}
		# --codonseparateonly option
		elsif (/^-+codonseparateonly=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$codonseparateonly = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$codonseparateonly = 0;
			}
			else {
				&errorMessage(__LINE__, 'Codon position dividing option is not valid.');
			}
		}
		# -w option
		elsif (/^-+(?:calcwhole|w)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$calcwhole = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$calcwhole = 0;
			}
			else {
				&errorMessage(__LINE__, 'The "calcwhole" option is not valid.');
			}
		}
		# -f option
		elsif (/^-+(?:basefreq|f)=(.+)/i) {
			if ($1 =~ /^Empirical$/i) {
				$basefreq = 'Empirical';
			}
			elsif ($1 =~ /^Estimate$/i) {
				$basefreq = 'Estimate';
			}
			else {
				&errorMessage(__LINE__, 'BaseFreq option is not valid.');
			}
		}
		# -p option
		elsif (/^-+(?:partition|p)=(.+)/i) {
			foreach my $partition (split(/,/, $1)) {
				if ($partition =~ s/^([^:]+)://) {
					my $partname = $1;
					my @range;
					if ($partname eq 'whole') {
						&errorMessage(__LINE__, 'Partition name "whole" is reserved by Kakusan.');
					}
					elsif ($partname =~ /(_P[1-3])/i) {
						&errorMessage(__LINE__, "Partition name \"-$1\" is reserved by Kakusan.");
					}
					unless ($part2site{$partname}) {
						push(@partname, $partname);
					}
					else {
						&errorMessage(__LINE__, "Partition name \"$partname\" is doubly specified.");
					}
					# case of normal partition
					if ($partition =~ /^(\d+)-(\d+)$/) {
						@range = &range2list($1, $2, 1);
					}
					else {
						&errorMessage(__LINE__, 'Partition specification is not valid.');
					}
					# Set site numbers to partition name
					foreach my $site (@range) {
						unless ($site2part{$site}) {
							$site2part{$site} = $partname;
						}
						else {
							&errorMessage(__LINE__, "Site $site is overlapping between \"$site2part{$site}\" and \"$partname\".");
						}
						push(@{$part2site{$partname}}, $site);
					}
				}
				else {
					&errorMessage(__LINE__, 'Partition name is not specified.');
				}
			}
		}
		# -g option
		elsif (/^-+(?:ngammacat|g)=(.+)/i) {
			if ($1 =~ /^(\d+)$/) {
				if ($1 < 2) {
					&errorMessage(__LINE__, 'The number of rate categories is not valid.');
				}
				else {
					$ngammacat = $1;
				}
			}
			else {
				&errorMessage(__LINE__, 'The number of rate categories is not an integer.');
			}
		}
		# -m option
		elsif (/^-+(?:ngam|m)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$n_gam = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$n_gam = 0;
			}
			else {
				&errorMessage(__LINE__, 'The "ngam" option is not valid.');
			}
		}
		# -r option
		elsif (/^-+(?:agamma|r)=(.+)/i) {
			if ($1 =~ /^Enable$/i) {
				$agamma = 1;
			}
			elsif ($1 =~ /^Disable$/i) {
				$agamma = 0;
			}
			else {
				&errorMessage(__LINE__, 'The "agamma" option is not valid.');
			}
		}
		# -a option
		elsif (/^-+(?:algorithm|a)=(.+)/i) {
			if ($1 =~ /^NJJC$/i) {
				$algorithm = 'NJJC';
			}
			elsif ($1 =~ /^NJLogDet$/i) {
				$algorithm = 'NJLogDet';
			}
			elsif ($1 =~ /^SARandom$/i) {
				$algorithm = 'SARandom';
			}
			elsif ($1 =~ /^SAClosest$/i) {
				$algorithm = 'SAClosest';
			}
			elsif ($1 =~ /^SAFurthest$/i) {
				$algorithm = 'SAFurthest';
			}
			else {
				&errorMessage(__LINE__, 'Algorithm option of tree generation is not valid.');
			}
		}
		# -t option
		elsif (/^-+(?:tree|t)=(.+)/i) {
			if ($1 =~ /^Different$/i) {
				$difftreeeachpart = 1;
			}
			elsif ($1 =~ /^Common$/i) {
				next;
			}
			else {
				&errorMessage(__LINE__, 'Tree option for parameter optimization is not valid.');
			}
		}
		# -o option
		elsif (/^-+(?:output|o)=(.+)/i) {
			foreach (split(/,/, $1)) {
				if (/^MrBayes$/i) {
					$output{'mrbayes'} = 1;
				}
				elsif (/^PAUP$/i) {
					$output{'paup'} = 1;
				}
				elsif (/^Treefinder$/i) {
					$output{'treefinder'} = 1;
				}
				elsif (/^all$/i) {
					$output{'all'} = 1;
				}
				else {
					&errorMessage(__LINE__, 'Specified output option is not valid.');
				}
			}
		}
		# -n option
		elsif (/^-+(?:numthreads|n)=(.+)/i) {
			if ($1 =~ /^(\d+)$/) {
				if ($1 == 0) {
					&errorMessage(__LINE__, 'The number of threads is not valid.');
				}
				else {
					$numthreads = $1;
				}
			}
			else {
				&errorMessage(__LINE__, 'The number of threads is not an integer.');
			}
		}
		# -l option
		elsif (/^-+(?:likelihood|l)=(.+)/i) {
			if ($1 =~ /^BASEML$/i) {
				$lnlcalc = 'baseml';
			}
			elsif ($1 =~ /^PAUP$/i) {
				$lnlcalc = 'paup';
			}
			elsif ($1 =~ /^TF$/i) {
				$lnlcalc = 'tf';
			}
			else {
				&errorMessage(__LINE__, 'Likelihood calculator option is not valid.');
			}
		}
		# --treefile option
		elsif (/^-+treefile=(.+)/i) {
			if (-e "$1") {
				$treefile = $1;
			}
			else {
				&errorMessage(__LINE__, "Specified tree file \"$1\" does not exist.");
			}
		}
		# input file
		else {
			if (!$infile{$_} && -f $_) {
				$infile{$_} = 1;
				push(@infile, $_);
			}
			elsif ($infile{$_}) {
				&errorMessage(__LINE__, "The input file \"$_\" is doubly specified.");
			}
			else {
				&errorMessage(__LINE__, "Specified input file \"$_\" does not exist or this is not a file.");
			}
		}
	}
	# Is the input file specified?
	unless (%infile) {
		print("No input files are specified.\nEntering interactive mode.\nSpecified options are ignored.\nSpecify an input file name.\n");
		$interactive = 1;
		my $response = 1;
		while ($response) {
			chomp($response = <STDIN>);
			$response =~ s/^\s*(.+?)\s*$/$1/;
			$response =~ s/^"(.+)"$/$1/;
			unless ($response) {
				if (%infile) {
					print("OK. Input file specification have terminated.\n\n");
					last;
				}
				else {
					$response = 1;
					print("Input is empty.\nSpecify an input file name.\n");
				}
			}
			elsif (!$infile{$response} && -f $response) {
				$infile{$response} = 1;
				push(@infile, $response);
				print("OK. An input file was accepted.\nSpecify an input file name or just press enter.\n");
			}
			elsif ($infile{$response}) {
				print("The input file is already accepted.\nSpecify an input file name or just press enter.\n");
			}
			elsif (!-f $response) {
				if (%infile) {
					print("The input file does not exist.\nSpecify an input file name or just press enter.\n");
				}
				else {
					print("The input file does not exist.\nSpecify an input file name.\n");
				}
			}
		}
	}
	undef(%infile);
	my $datadir;
	my $coding;
	foreach (@infile) {
		my $fullpathname = File::Spec->rel2abs($_);
		push(@fullpathname, $fullpathname);
		my ($volume, $directories, $file) = File::Spec->splitpath($fullpathname);
		push(@inputfile, $file);
		if ($file =~ /_P(?:\.[^\.]*)?$/i) {
			$coding ++;
		}
		if (@fullpathname == 1) {
			$datadir = $volume . $directories;
		}
	}
	$maindir = $datadir . $inputfile[0] . '.kakusan';
	print("Log, result and configuration files are output to \"$maindir\".\n\n");
	if ($interactive) {
		print("OUTPUT OPTIONS\n\n");
		my $loop = 1;
		while ($loop) {
			my $response;
			print("Do you want the model configuration files for PAUP*? (y/n)\n(default: y)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*(?:yes|y)\s*$/i || $response =~ /^\s*$/) {
				$output{'paup'} = 1;
				print("OK. The model configuration files for PAUP* will be output.\n\n");
				last;
			}
			elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
				print("OK. The model configuration files for PAUP* will NOT be output.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		while ($loop) {
			my $response;
			print("Do you want the model configuration files for MrBayes? (y/n)\n(default: y)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*(?:yes|y)\s*$/i || $response =~ /^\s*$/) {
				$output{'mrbayes'} = 1;
				print("OK. The model configuration files for MrBayes will be output.\n\n");
				last;
			}
			elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
				print("OK. The model configuration files for MrBayes will NOT be output.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		while ($loop) {
			my $response;
			print("Do you want the model configuration files for Treefinder? (y/n)\n(default: y)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*(?:yes|y)\s*$/i || $response =~ /^\s*$/) {
				$output{'treefinder'} = 1;
				print("OK. The model configuration files for Treefinder will be output.\n\n");
				last;
			}
			elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
				print("OK. The model configuration files for Treefinder will NOT be output.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		if (!$output{'mrbayes'} && !$output{'paup'} && !$output{'treefinder'}) {
			&errorMessage(__LINE__, 'All output options are disabled.');
		}
		print("ANALYSIS OPTIONS\n\n");
		if ($coding && $output{'treefinder'} != 1) {
			while ($loop) {
				my $response;
				print("You input protein coding sequence.\nDo you want to run separated analyses of each codon position? (y/n)\n(default: n)\n");
				chomp($response = <STDIN>);
				if ($response =~ /^\s*(?:yes|y)\s*$/i) {
					$dividecodon = 1;
					print("OK. Separated analyses of each codon position is enabled.\n\n");
					last;
				}
				elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
					$dividecodon = 0;
					print("OK. Separated analyses of each codon position is disabled.\n\n");
					last;
				}
				else {
					print("Cannot understand your answer.\n\n");
				}
			}
		}
		elsif ($coding && $output{'treefinder'} == 1) {
			$dividecodon = 1;
		}
		if ($dividecodon == 1) {
			while ($loop) {
				my $response;
				print("You enabled separated analyses of each codon position.\nDo you want to disable codon position combined analyses of each partition? (y/n)\n(default: n)\n");
				chomp($response = <STDIN>);
				if ($response =~ /^\s*(?:yes|y)\s*$/i) {
					$codonseparateonly = 1;
					if (@fullpathname == 1) {
						$calcwhole = 0;
					}
					print("OK. Codon position combined analyses of each partition is disabled.\n\n");
					last;
				}
				elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
					$codonseparateonly = 0;
					if (@fullpathname == 1) {
						$calcwhole = 1;
					}
					print("OK. Codon position combined analyses of each partition is enabled.\n\n");
					last;
				}
				else {
					print("Cannot understand your answer.\n\n");
				}
			}
		}
		if (@fullpathname > 1 && $output{'paup'} != 1) {
			while ($loop) {
				my $response;
				print("You input multiple files.\nDo you want to run combined analysis? (y/n)\n(default: n)\n");
				chomp($response = <STDIN>);
				if ($response =~ /^\s*(?:yes|y)\s*$/i) {
					$calcwhole = 1;
					print("OK. Combined analysis is enabled.\n\n");
					last;
				}
				elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
					$calcwhole = 0;
					print("OK. Combined analysis is disabled.\n\n");
					last;
				}
				else {
					print("Cannot understand your answer.\n\n");
				}
			}
		}
		elsif (@fullpathname > 1 && $output{'paup'} == 1) {
			$calcwhole = 1;
		}
		while ($loop) {
			my $response;
			print("Which do you want to use the program for likelihood calculation? (baseml/tf/paup)\n(default: baseml)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*paup\s*$/i) {
				$lnlcalc = 'paup';
				print("OK. The likelihoods will be calculated by PAUP.\n\n");
				last;
			}
			elsif ($response =~ /^\s*tf\s*$/i) {
				$lnlcalc = 'tf';
				print("OK. The likelihoods will be calculated by TF.\n\n");
				last;
			}
			elsif ($response =~ /^\s*baseml\s*$/i || $response =~ /^\s*$/) {
				$lnlcalc = 'baseml';
				print("OK. The likelihoods will be calculated by BASEML.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		while ($loop) {
			my $response;
			print("Do you want to use chi-square test for pre-selection sorting of the candidate models? (y/n)\n(default: y)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*(?:yes|y)\s*$/i || $response =~ /^\s*$/) {
				$chisq = 1;
				print("OK. Pre-selection model sorting is enabled.\n\n");
				last;
			}
			elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
				$chisq = 0;
				print("OK. Pre-selection model sorting is disabled.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		while ($loop) {
			my $response;
			print("Do you want to optimize the parameters of base composition? (y/n)\n(default: y)\n");
			chomp($response = <STDIN>);
			if ($response =~ /^\s*(?:yes|y)\s*$/i || $response =~ /^\s*$/) {
				$basefreq = 'Estimate';
				print("OK. The parameters will be optimized.\n\n");
				last;
			}
			elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
				$basefreq = 'Empirical';
				print("OK. The empirical values will be used.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		while ($loop) {
			my $response;
			print("How many rate categories do you want to consider on discrete gamma rate heterogeneity? (integer)\n(default: 8)\n");
			chomp($response = <STDIN>);
			unless ($response) { $response = 8; }
			if ($response =~ /^(\d+)$/) {
				if ($1 < 2) {
					&errorMessage(__LINE__, 'The number of rate categories is not valid.');
				}
				else {
					$ngammacat = $1;
				}
				print("OK. The number of rate categories is set to $1.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
		if ($lnlcalc eq 'baseml') {
			if ($coding || @fullpathname > 1) {
				while ($loop) {
					my $response;
					print("Do you want to consider N-GAM heterogeneity? (y/n)\nNote that this model is very time-consuming.\n(default: n)\n");
					chomp($response = <STDIN>);
					if ($response =~ /^\s*(?:yes|y)\s*$/i) {
						$n_gam = 1;
						print("OK. N-GAM heterogeneity is enabled.\n\n");
						last;
					}
					elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
						$n_gam = 0;
						print("OK. N-GAM heterogeneity is disabled.\n\n");
						last;
					}
					else {
						print("Cannot understand your answer.\n\n");
					}
				}
			}
			while ($loop) {
				my $response;
				print("Do you want to consider autocorrelated discrete gamma heterogeneity? (y/n)\nNote that this model is very time-consuming.\n(default: n)\n");
				chomp($response = <STDIN>);
				if ($response =~ /^\s*(?:yes|y)\s*$/i) {
					$agamma = 1;
					print("OK. Autocorrelated discrete gamma heterogeneity is enabled.\n\n");
					last;
				}
				elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
					$agamma = 0;
					print("OK. Autocorrelated discrete gamma heterogeneity is disabled.\n\n");
					last;
				}
				else {
					print("Cannot understand your answer.\n\n");
				}
			}
		}
		if (@fullpathname > 1) {
			while ($loop) {
				my $response;
				print("Do you want to use different tree topologies for parameter optimization on each partiton? (y/n)\n(default: n)\n");
				chomp($response = <STDIN>);
				if ($response =~ /^\s*(?:yes|y)\s*$/i) {
					$difftreeeachpart = 1;
					print("OK. Different tree topologies will be used for parameter optimization on each partiton.\n\n");
					last;
				}
				elsif ($response =~ /^\s*(?:no|n)\s*$/i || $response =~ /^\s*$/) {
					$difftreeeachpart = 0;
					print("OK. A common tree topology will be used for parameter optimization on each partiton.\n\n");
					last;
				}
				else {
					print("Cannot understand your answer.\n\n");
				}
			}
		}
		while ($loop) {
			my $response;
			print("How many processes do you want to run simultaneously? (integer)\n(default: 1)\n");
			chomp($response = <STDIN>);
			unless ($response) { $response = 1; }
			if ($response =~ /^\s*(\d+)\s*$/) {
				if ($1 == 0) {
					&errorMessage(__LINE__, 'The number of threads is not valid.');
				}
				else {
					$numthreads = $1;
				}
				print("OK. The number of processes is set to $1.\n\n");
				last;
			}
			else {
				print("Cannot understand your answer.\n\n");
			}
		}
	}
	# Validation
	if (@partname && @inputfile > 1) {
		&errorMessage(__LINE__, 'Multiple input files and command line partition specification is exclusive.');
	}
	elsif (!@partname) {
		my %temphash;
		foreach (@inputfile) {
			$_ =~ s/ /_/g;
			$_ =~ /^(.+)\.[^\.]*$/;
			my $partname = $1;
			if ($partname eq 'whole') {
				&errorMessage(__LINE__, 'Partition name "whole" is reserved by Kakusan.');
			}
			elsif ($partname =~ /(_P[1-3])$/i) {
				&errorMessage(__LINE__, "Partition name \"-$1\" is reserved by Kakusan.");
			}
			unless ($temphash{$partname}) {
				$temphash{$partname} = 1;
				push(@partname, $partname);
			}
			else {
				&errorMessage(__LINE__, "Partition name \"$partname\" is doubly specified.");
			}
		}
	}
	# calcwhole option validation
	if ($calcwhole == 0 && @partname < 2 && ($partname[0] !~ /_P$/i || $partname[0] =~ /_P$/i && !$dividecodon)) {
		$calcwhole = 1;
	}
	# dividecodon option validation
	if ($dividecodon) {
		my $coding;
		foreach (@partname) {
			if (/_P$/i) {
				$coding = 1;
			}
		}
		unless ($coding) {
			&errorMessage(__LINE__, 'Codon position dividing option is enabled. But coding sequence is not input.');
		}
	}
	else {
		if ($codonseparateonly) {
			&errorMessage(__LINE__, 'Codon position dividing option is disabled. But codon position combined analyses are disabled.');
		}
	}
	# Are there unused sites?
	for (my $i = 1; $i <= keys(%site2part); $i ++) {
		unless ($site2part{$i}) {
			&errorMessage(__LINE__, 'Partition specification is not valid. There are unused sites.');
		}
	}
	# If output option is not specified...
	if (!%output || $output{'mrbayes'} && $output{'paup'} && $output{'treefinder'}) {
		undef(%output);
		$output{'all'} = 1;
	}
	# Output option processing
	if ($output{'all'}) {
		foreach (@allmodel) {
			$model{$_} = 1;
		}
		foreach (@allheterogeneity) {
			$heterogeneity{$_} = 1;
		}
	}
	if ($output{'mrbayes'}) {
		foreach (@modelformrbayes) {
			$model{$_} = 1;
		}
		foreach (@heterogeneityformrbayes) {
			$heterogeneity{$_} = 1;
		}
	}
	if ($output{'paup'}) {
		foreach (@modelforpaup) {
			$model{$_} = 1;
		}
		foreach (@heterogeneityforpaup) {
			$heterogeneity{$_} = 1;
		}
	}
	if ($output{'treefinder'}) {
		foreach (@modelfortf) {
			$model{$_} = 1;
		}
		foreach (@heterogeneityfortf) {
			$heterogeneity{$_} = 1;
		}
	}
	# If N-GAM disabled
	unless ($n_gam) {
		delete($heterogeneity{'CodonPos3Gamma'});
		delete($heterogeneity{'CodonPos3AGamma'});
		delete($heterogeneity{'GeneNGamma'});
		delete($heterogeneity{'GeneNAGamma'});
		delete($heterogeneity{'GeneCodonPosNGamma'});
		delete($heterogeneity{'GeneCodonPosNAGamma'});
	}
	# If AGamma disabled
	unless ($agamma) {
		delete($heterogeneity{'AGamma'});
		delete($heterogeneity{'CodonPos1AGamma'});
		delete($heterogeneity{'CodonPos3AGamma'});
		delete($heterogeneity{'Gene1AGamma'});
		delete($heterogeneity{'GeneNAGamma'});
		delete($heterogeneity{'GeneCodonPos1AGamma'});
		delete($heterogeneity{'GeneCodonPosNAGamma'});
	}
	# if BASEML or TF mode
	if ($lnlcalc ne 'paup' && $algorithm =~ /^SA/) {
		&errorMessage(__LINE__, 'Cannot use stepwise addition in BASEML/TF mode.');
	}
	#
	print("done.\n\n");
#	# Output to screen
#	my $lctime = localtime();
#	print <<"_END";
#Time: $lctime
#Input File Name: $inputfile
#Chi-Square Test: $chisq
#Significant Level: $siglevel
#Calc whole: $calcwhole
#Divide codon: $dividecodon
#Base Frequencies: $basefreq
#Rate Categories: $ngammacat
#Algorithm: $algorithm
#Different Tree on each Partition: $difftreeeachpart
#Threads: $numthreads
#Partitions: @partname
#_END
#	foreach my $partname (@partname) {
#		print("$partname: @{$part2site{$partname}}\n");
#	}
#	print("Matrix:");
#	foreach my $model (keys(%model)) {
#		print(" $model");
#	}
#	print("\n");
#	print("Heterogeneity:");
#	foreach my $heterogeneity (keys(%heterogeneity)) {
#		print(" $heterogeneity");
#	}
#	print("\n");
#	print("\n");
}

sub range2list {
	# Input: Site of range start, Site of range end, Skip number of sites
	# Output: List of sites which belong to the range
	my ($start, $end, $skip) = @_;
	my @num;
	if ($start != 0 && $skip != 0 && $start <= $end && $skip <= $end - $start) {
		for (my $i = $start; $i <= $end; $i += $skip) {
			push(@num, $i);
		}
	}
	else {
		&errorMessage(__LINE__, 'Partition specification is not valid.');
	}
	return(@num);
}

sub checkFiles {
	print("Checking commands and files...\n");
	# Check working directory
	my $datadir;
	{
		my $maindirname;
		{
			my ($volume, $directories, $file) = File::Spec->splitpath($maindir);
			$datadir = $volume . $directories;
			$maindirname = $file;
		}
		my $tempno;
		while (-e $maindir) {
			$tempno ++;
			unless (-e $datadir . "bkup$tempno.$maindirname") {
				my $loop = 1;
				while ($loop) {
					print("\"$maindir\" already exists. Can I rename it to \"$datadir" . "bkup$tempno.$maindirname\"?\n");
					my $response;
					chomp($response = <STDIN>);
					if ($response =~ /^\s*(?:yes|y)\s*$/i) {
						unless (rename($maindir, $datadir . "bkup$tempno.$maindirname")) {
							&errorMessage(__LINE__, 'Cannot change the directory name.');
						}
						else {
							print("OK. Changed to \"$datadir" . "bkup$tempno.$maindirname\".\n");
							last;
						}
					}
					elsif ($response =~ /^\s*(?:no|n)\s*$/i) {
						&errorMessage(__LINE__, 'The script was aborted.');
					}
					else {
						print("Cannot understand your answer.\n\n");
					}
				}
			}
		}
	}
	# Make working directory
	my $kakusanpath;
	BEGIN {
		use File::Spec;
		my ($volume, $directories, $file) = File::Spec->splitpath(File::Spec->rel2abs($0));
		$kakusanpath = $volume . $directories;
	}
	use lib qq($kakusanpath);
	use Cwd;
	my $devnull = File::Spec->devnull();
	my $currentdir;
	{
		my ($volume, $directories, $file) = File::Spec->splitpath(File::Spec->catfile(Cwd->cwd(), 'tmp'));
		$currentdir = $volume . $directories;
	}
	unless (mkdir($maindir)) {
		&errorMessage(__LINE__, "Cannot make \"$maindir\" directory.");
	}
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	unless (mkdir('Logs')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/Logs\" directory.");
	}
	# Check ReadSeq
	if (system('java -jar "' . $currentdir . 'readseq.jar" -h 2> ' . $devnull . ' 1> ' . $devnull)) {
		if (system('java -jar "' . $datadir . 'readseq.jar" -h 2> ' . $devnull . ' 1> ' . $devnull)) {
			if (system('java -jar "' . $kakusanpath . 'readseq.jar" -h 2> ' . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$currentdir$readseq\" -h 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("\"$datadir$readseq\" -h 2> " . $devnull . ' 1> ' . $devnull)) {
						if (system("\"$kakusanpath$readseq\" -h 2> " . $devnull . ' 1> ' . $devnull)) {
							if (system("$readseq -h 2> " . $devnull . ' 1> ' . $devnull)) {
								&errorMessage(__LINE__, 'Cannot use readseq. Please install readseq.');
							}
						}
						else {
							$readseq = "\"$kakusanpath$readseq\"";
						}
					}
					else {
						$readseq = "\"$datadir$readseq\"";
					}
				}
				else {
					$readseq = "\"$currentdir$readseq\"";
				}
			}
			else {
				$readseq = 'java -jar "' . $kakusanpath . 'readseq.jar"';
			}
		}
		else {
			$readseq = 'java -jar "' . $datadir . 'readseq.jar"';
		}
	}
	else {
		$readseq = 'java -jar "' . $currentdir . 'readseq.jar"';
	}
	# Check PAUP
	unless (open(NEXUS, '> NEXUS.nex')) {
		&errorMessage(__LINE__, 'Cannot make file.');
	}
	print(NEXUS "#NEXUS\nBegin PAUP;\nQuit;\nEnd;\n");
	close(NEXUS);
	if (system("\"$currentdir$paupcmd\" -u -n NEXUS.nex 2> " . $devnull . ' 1> ' . $devnull)) {
		if (system("\"$datadir$paupcmd\" -u -n NEXUS.nex 2> " . $devnull . ' 1> ' . $devnull)) {
			if (system("\"$kakusanpath$paupcmd\" -u -n NEXUS.nex 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("$paupcmd -u -n NEXUS.nex 2> " . $devnull . ' 1> ' . $devnull)) {
					print("Cannot find PAUP*.\n");
					undef($paupcmd);
					if ($lnlcalc eq 'paup') {
							&errorMessage(__LINE__, 'The analysis is terminated.');
					}
				}
			}
			else {
				$paupcmd = "\"$kakusanpath$paupcmd\"";
			}
		}
		else {
			$paupcmd = "\"$datadir$paupcmd\"";
		}
	}
	else {
		$paupcmd = "\"$currentdir$paupcmd\"";
	}
	unlink('NEXUS.nex');
	if ($n_gam && $lnlcalc eq 'paup') {
		&errorMessage(__LINE__, "Cannot use NGamma in PAUP mode.");
	}
	if ($agamma && $lnlcalc eq 'paup') {
		&errorMessage(__LINE__, "Cannot use AGamma in PAUP mode.");
	}
	# Check PHYLIP & BASEML/TF
	if ($lnlcalc ne 'paup') {
		unless (open(SEQ, '> tempseq.phy')) {
			&errorMessage(__LINE__, 'Cannot make file.');
		}
		print(SEQ " 3 12\nTax1      ACGTACGTACGT\nTax2      AGCTAGCTAGCT\nTax3      ATCGATCGATCG\n");
		close(SEQ);
		unless (open(CMD, '> command')) {
			&errorMessage(__LINE__, 'Cannot make file.');
		}
		print(CMD "tempseq.phy\nY\n");
		close(CMD);
		if (system("\"$currentdir$dnadist\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
			if (system("\"$datadir$dnadist\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$kakusanpath$dnadist\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("$dnadist < command 2> " . $devnull . ' 1> ' . $devnull)) {
						&errorMessage(__LINE__, 'Cannot run $dnadist. Please check $dnadist.');
					}
				}
				else {
					$dnadist = "\"$kakusanpath$dnadist\"";
				}
			}
			else {
				$dnadist = "\"$datadir$dnadist\"";
			}
		}
		else {
			$dnadist = "\"$currentdir$dnadist\"";
		}
		rename('outfile','tempdist.phy');
		if (system("\"$currentdir$dnapars\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
			if (system("\"$datadir$dnapars\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$kakusanpath$dnapars\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("$dnapars < command 2> " . $devnull . ' 1> ' . $devnull)) {
						&errorMessage(__LINE__, 'Cannot run $dnapars. Please check $dnapars.');
					}
					else {
						unlink('outtree');
						unlink('outfile');
					}
				}
				else {
					unlink('outtree');
					unlink('outfile');
					$dnapars = "\"$kakusanpath$dnapars\"";
				}
			}
			else {
				unlink('outtree');
				unlink('outfile');
				$dnapars = "\"$datadir$dnapars\"";
			}
		}
		else {
			unlink('outtree');
			unlink('outfile');
			$dnapars = "\"$currentdir$dnapars\"";
		}
		unless (open(CMD, '> command')) {
			&errorMessage(__LINE__, 'Cannot make file.');
		}
		print(CMD "tempdist.phy\nY\n");
		close(CMD);
		if (system("\"$currentdir$neighbor\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
			if (system("\"$datadir$neighbor\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$kakusanpath$neighbor\" < command 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("$neighbor < command 2> " . $devnull . ' 1> ' . $devnull)) {
						&errorMessage(__LINE__, 'Cannot run $neighbor. Please check $neighbor.');
					}
				}
				else {
					$neighbor = "\"$kakusanpath$neighbor\"";
				}
			}
			else {
				$neighbor = "\"$datadir$neighbor\"";
			}
		}
		else {
			$neighbor = "\"$currentdir$neighbor\"";
		}
		unlink('outfile');
		unlink('command');
		unlink('tempdist.phy');
		if ($lnlcalc eq 'baseml') {
			unless (open(CMD, '> baseml.ctl')) {
				&errorMessage(__LINE__, 'Cannot make file.');
			}
			print(CMD "seqfile = tempseq.phy\ntreefile = outtree\noutfile = outfile\n");
			close(CMD);
			if (system("\"$currentdir$baseml\" 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$datadir$baseml\" 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("\"$kakusanpath$baseml\" 2> " . $devnull . ' 1> ' . $devnull)) {
						if (system("$baseml 2> " . $devnull . ' 1> ' . $devnull)) {
							&errorMessage(__LINE__, 'Cannot run $baseml. Please check $baseml.');
						}
					}
					else {
						$baseml = "\"$kakusanpath$baseml\"";
					}
				}
				else {
					$baseml = "\"$datadir$baseml\"";
				}
			}
			else {
				$baseml = "\"$currentdir$baseml\"";
			}
			unlink('outfile');
			unlink('2base.t');
			unlink('lnf');
			unlink('rst');
			unlink('rst1');
			unlink('rub');
			unlink('baseml.ctl');
		}
		elsif ($lnlcalc eq 'tf') {
			unless (open(CMD, '> quit.tl')) {
				&errorMessage(__LINE__, 'Cannot make file.');
			}
			print(CMD "Quit\n");
			close(CMD);
			if (system("\"$currentdir$tf\" quit.tl 2> " . $devnull . ' 1> ' . $devnull)) {
				if (system("\"$datadir$tf\" quit.tl 2> " . $devnull . ' 1> ' . $devnull)) {
					if (system("\"$kakusanpath$tf\" quit.tl 2> " . $devnull . ' 1> ' . $devnull)) {
						if (system("$tf quit.tl 2> " . $devnull . ' 1> ' . $devnull)) {
							&errorMessage(__LINE__, "Cannot run $tf. Please check $tf.");
						}
					}
					else {
						$tf = "\"$kakusanpath$tf\"";
					}
				}
				else {
					$tf = "\"$datadir$tf\"";
				}
			}
			else {
				$tf = "\"$currentdir$tf\"";
			}
			unlink('quit.tl');
		}
		unlink('outtree');
		unlink('tempseq.phy');
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub convertSequences {
	print("Converting sequence files...\n");
	push(@allpart, 'whole');
	for (my $i = 0; $i < @inputfile; $i ++) {
		# Preprocess input file
		unless (open(INPUT, "< $fullpathname[$i]")) {
			&errorMessage(__LINE__, "Cannot open \"$fullpathname[$i]\".");
		}
		my @seqdata;
		while (<INPUT>) {
			push(@seqdata, split(/(?:\r\n|\r|\n)/));
		}
		close(INPUT);
		unless (chdir($maindir)) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		unless (open(OUTPUT, "> $inputfile[$i]")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/$inputfile[$i]\".");
		}
		my $nexusornot;
		my $tempntax;
		my $tempnchar;
		foreach (@seqdata) {
			if (/^#NEXUS\s*$/i) {
				print(OUTPUT "$_\n");
				$nexusornot = 1;
			}
			if ($nexusornot == 1 && (/Begin\s+Data\s*;/i || /Begin\s+Characters\s*;/i)) {
				$nexusornot = 2;
			}
			if ($nexusornot == 2 && /NTax=(\d+)/i) {
				$tempntax = $1;
			}
			if ($nexusornot == 2 && /NChar=(\d+)/i) {
				$tempnchar = $1;
			}
			if ($nexusornot == 2 && /End\s*;/i) {
				print(OUTPUT "$_\n");
				$nexusornot = 1;
			}
			if ($nexusornot != 1) {
				print(OUTPUT "$_\n");
			}
		}
		if ($nexusornot && $tempntax && $tempnchar && $paupcmd) {
			print(OUTPUT "Begin PAUP;\n\tExport File=$inputfile[$i]" . "_temp.nex Format=NEXUS;\n\tQuit;\nEnd;\n");
		}
		close(OUTPUT);
		my $outpart;
		if (@inputfile == 1) {
			$outpart = 'whole';
		}
		else {
			$outpart = $partname[$i];
			push(@allpart, $partname[$i]);
		}
		if ($nexusornot && $tempntax && $tempnchar && $paupcmd) {
			if (system("$paupcmd $inputfile[$i] > Logs/convert_$inputfile[$i].txt")) {
				&errorMessage(__LINE__, "Cannot run $paupcmd. Please check $paupcmd.");
			}
			unless (open(TEMP, "< $inputfile[$i]" . "_temp.nex")) {
				&errorMessage(__LINE__, "Cannot open \"$maindir/$inputfile[$i]" . '_temp.nex".');
			}
			unless (open(OUTPUT, "> data_$outpart.nex")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_$outpart.nex\".");
			}
			while (<TEMP>) {
				s/(datatype=dna)[^\r\n]+/$1 gap=- missing=?;/i;
				print(OUTPUT $_);
			}
			close(OUTPUT);
			close(TEMP);
			unlink("$inputfile[$i]" . "_temp.nex");
		}
		else {
			# Convert to FASTA by ReadSeq
			if (system("$readseq -a -C -f8 -o=$inputfile[$i].fas $inputfile[$i]")) {
				&errorMessage(__LINE__, 'Cannot convert sequence file.');
			}
			# Convert ReadSeq output FASTA to NEXUS
			unless (open(FASTA, "< $inputfile[$i].fas")) {
				&errorMessage(__LINE__, "Cannot open \"$maindir/$inputfile[$i].fas\".");
			}
			$tempntax = 0;
			while (<FASTA>) {
				if (/^>[^\s,]+.*?,? (\d+) .+\n$/) {
					$tempntax ++;
					unless ($tempnchar) {
						$tempnchar = $1;
					}
					else {
						if ($tempnchar != $1) {
							&errorMessage(__LINE__, 'Sequence lengths are not equal.');
						}
					}
				}
			}
			close(FASTA);
			if ($tempntax < 3) {
				&errorMessage(__LINE__, 'The number of taxon is too small or file parsing error.');
			}
			unless (open(NEXUS, "> data_$outpart.nex")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_$outpart.nex\".");
			}
			print(NEXUS '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$tempntax NChar=$tempnchar;\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix");
			unless (open(FASTA, "< $inputfile[$i].fas")) {
				&errorMessage(__LINE__, "Cannot open \"$maindir/$inputfile[$i].fas\".");
			}
			while (<FASTA>) {
				s/^>([^\s,]+).*?,? \d+ .+\r?\n?$/\n$1 /;
				s/^([^>][^\r\n]*?)\r?\n?$/$1/;
				print(NEXUS $_);
			}
			close(FASTA);
			print(NEXUS "\n\t;\nEnd;\n");
			close(NEXUS);
		}
		# Output PHYLIP format
		if ($lnlcalc ne 'paup') {
			unless (open(PHYLIP, "> data_$outpart.phy")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_$outpart.phy\".");
			}
			print(PHYLIP " $tempntax $tempnchar\n");
			unless (open(NEXUS, "< data_$outpart.nex")) {
				&errorMessage(__LINE__, "Cannot open \"$maindir/data_$outpart.nex\".");
			}
			my $line = 0;
			my @temptaxa;
			while (<NEXUS>) {
				$line ++;
				if ($line < 7) {
					next;
				}
				elsif ($line >= 7 && $line < 7 + $tempntax) {
					if (/^(\S+)\s+(\S+)\r?\n?$/) {
						push(@temptaxa, $1);
						printf(PHYLIP "Tax%-5d  $2\n", $line - 6);
					}
					else {
						&errorMessage(__LINE__, "\"$maindir/data_$outpart.nex\" is not valid.");
					}
				}
				else {
					last;
				}
			}
			close(NEXUS);
			close(PHYLIP);
			# Output taxon list
			unless (-e "taxonlist.txt") {
				unless (open(TXT, "> taxonlist.txt")) {
					&errorMessage(__LINE__, "Cannot make \"$maindir/taxonlist.txt\".");
				}
				for (my $i = 0; $i < $tempntax; $i ++) {
					print(TXT 'Tax' . ($i + 1) . "\t" . $temptaxa[$i] . "\n");
				}
				close(TXT);
			}
		}
		unless ($ntax) {
			$ntax = $tempntax;
		}
		if ($ntax != $tempntax) {
			&errorMessage(__LINE__, 'Invalid input files. The number of taxa is not equal.');
		}
		if ($outpart eq 'whole' && @partname == 1 && !$part2site{$partname[0]}) {
			push(@{$part2site{$partname[0]}}, &range2list(1, $tempnchar, 1));
			$nchar = $tempnchar;
		}
		elsif ($outpart eq 'whole') {
			$nchar = $tempnchar;
		}
		else {
			push(@{$part2site{$outpart}}, &range2list($nchar + 1, $nchar + $tempnchar, 1));
			$nchar += $tempnchar;
		}
		unless (chdir('..')) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
	}
	print("done.\n\n");
}

sub splitSequences {
	print("Splitting sequences...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	# Read sequence file and output to separated files
	unless (open(NEXUS, "< data_whole.nex")) {
		&errorMessage(__LINE__, "Cannot open \"$maindir/data_whole.nex\".");
	}
	my $line = 0;
	while (<NEXUS>) {
		$line ++;
		if ($line < 7) {
			next;
		}
		elsif ($line >= 7 && $line < 7 + $ntax) {
			if (/^(\S+)\s+(\S+)\r?\n?$/) {
				push(@taxa, $1);
				for (my $i = 0; $i < $nchar; $i ++) {
					push(@{$seqs{$1}}, uc(substr($2, $i, 1)));
				}
			}
			else {
				&errorMessage(__LINE__, "\"$maindir/data_whole.nex\" is not valid.");
			}
		}
		else {
			last;
		}
	}
	close(NEXUS);
	&calcTaxLength();
	if (@partname > 1) {
		foreach my $partname (@partname) {
			push(@allpart, $partname);
			unless (open(OUT, "> data_$partname.nex")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_$partname.nex\".");
			}
			print(OUT '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=" . scalar(@{$part2site{$partname}}) . ";\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix\n");
			foreach my $taxon (@taxa) {
				printf(OUT "%-*s ", $taxlength, $taxon);
				foreach my $site (@{$part2site{$partname}}) {
					print(OUT $seqs{$taxon}[$site - 1]);
				}
				print(OUT "\n");
			}
			print(OUT "\t;\nEnd;\n");
			close(OUT);
			if ($lnlcalc ne 'paup') {
				unless (open(OUT, "> data_$partname.phy")) {
					&errorMessage(__LINE__, "Cannot make \"$maindir/data_$partname.phy\".");
				}
				print(OUT " $ntax " . scalar(@{$part2site{$partname}}) . "\n");
				for (my $i = 0; $i < @taxa; $i ++) {
					printf(OUT "Tax%-5d  ", $i + 1);
					foreach my $site (@{$part2site{$partname}}) {
						print(OUT $seqs{$taxa[$i]}[$site - 1]);
					}
					print(OUT "\n");
				}
				close(OUT);
			}
		}
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub combineSequences {
	print("Combining sequences...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	my %taxa;
	for (my $i = 0; $i < @partname; $i ++) {
		unless (open(NEXUS, "< data_$partname[$i].nex")) {
			&errorMessage(__LINE__, "Cannot open \"$maindir/data_$partname[$i].nex\".");
		}
		my $line = 0;
		while (<NEXUS>) {
			$line ++;
			if ($line < 7) {
				next;
			}
			elsif ($line >= 7 && $line < 7 + $ntax) {
				if (/^(\S+)\s+(\S+)\r?\n?$/) {
					if ($i == 0 && !exists($taxa{$1})) {
						push(@taxa, $1);
						$taxa{$1} = 1;
					}
					elsif ($i == 0 && exists($taxa{$1})) {
						&errorMessage(__LINE__, 'Duplicated taxon.');
					}
					elsif (!exists($taxa{$1})) {
						&errorMessage(__LINE__, 'Input files have different taxa set.');
					}
					for (my $j = 0; $j < scalar(@{$part2site{$partname[$i]}}); $j ++) {
						push(@{$seqs{$1}}, uc(substr($2, $j, 1)));
					}
				}
				else {
					&errorMessage(__LINE__, "\"$maindir/data_$partname[$i].nex\" is not valid.");
				}
			}
			else {
				last;
			}
		}
		close(NEXUS);
	}
	&calcTaxLength();
	unless (open(OUT, "> data_whole.nex")) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/data_whole.nex\".");
	}
	print(OUT '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=$nchar;\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix\n");
	foreach my $taxon (@taxa) {
		printf(OUT "%-*s ", $taxlength, $taxon);
		for (my $i = 0; $i < $nchar; $i ++) {
			print(OUT $seqs{$taxon}[$i]);
		}
		print(OUT "\n");
	}
	print(OUT "\t;\nEnd;\n");
	close(OUT);
	if ($lnlcalc ne 'paup') {
		unless (open(OUT, "> data_whole.phy")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/data_whole.phy\".");
		}
		print(OUT " $ntax $nchar\n");
		for (my $i = 0; $i < @taxa; $i ++) {
			printf(OUT "Tax%-5d  ", $i + 1);
			for (my $j = 0; $j < $nchar; $j ++) {
				print(OUT $seqs{$taxa[$i]}[$j]);
			}
			print(OUT "\n");
		}
		close(OUT);
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub splitCodingSequences {
	foreach my $partname (@partname) {
		if ($partname =~ /_P$/i) {
			for (my $i = 0; $i < 3; $i ++) {
				push(@subpart, $partname . ($i + 1));
				my @range = &range2list($part2site{$partname}[$i], $part2site{$partname}[-1], 3);
				foreach my $site (@range) {
					push(@{$subpart{$partname . ($i + 1)}}, $site);
				}
			}
		}
	}
}

sub calcTaxLength {
	foreach my $taxon (@taxa) {
		if (length($taxon) > $taxlength) {
			$taxlength = length($taxon);
		}
	}
}

sub makeEachCodonPositions {
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	foreach my $partname (@subpart) {
		push(@allpart, $partname);
		unless (open(OUT, "> data_$partname.nex")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/data_$partname.nex\".");
		}
		print(OUT '#NEXUS' . "\n\nBegin Data;\n\tDimensions NTax=$ntax NChar=" . scalar(@{$subpart{$partname}}) . ";\n\tFormat DataType=DNA Gap=- Missing=?;\n\tMatrix\n");
		foreach my $taxon (@taxa) {
			printf(OUT "%-*s ", $taxlength, $taxon);
			foreach my $site (@{$subpart{$partname}}) {
				print(OUT $seqs{$taxon}[$site - 1]);
			}
			print(OUT "\n");
		}
		print(OUT "\t;\nEnd;\n");
		close(OUT);
		if ($lnlcalc ne 'paup') {
			# output PHYLIP format
			unless (open(OUT, "> data_$partname.phy")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_$partname.phy\".");
			}
			print(OUT ' ' . $ntax . ' ' . scalar(@{$subpart{$partname}}) . "\n");
			for (my $i = 0; $i < @taxa; $i ++) {
				printf(OUT "Tax%-5d  ", $i + 1);
				foreach my $site (@{$subpart{$partname}}) {
					print(OUT $seqs{$taxa[$i]}[$site - 1]);
				}
				print(OUT "\n");
			}
			close(OUT);
		}
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
}

sub setPartitions {
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	if (@partname > 1) {
		foreach my $partname (@partname) {
			if ($partname =~ /_P$/i) {
				# Output NEXUS partitioning command for codon positions
				unless (open(NEXUS, "> part_$partname" . '_codonpos.nex')) {
					&errorMessage(__LINE__, "Cannot make \"$maindir/part_$partname" . '_codonpos.nex".');
				}
				print(NEXUS "#NEXUS\n");
				print(NEXUS "\nBegin PAUP;\n" . &part_codonpos('PAUP') . "End;\n");
				close(NEXUS);
				# output PAML format for codon positions
				if ($lnlcalc eq 'baseml') {
					unless (open(OUT, "> data_$partname" . '_codonpos.phy')) {
						&errorMessage(__LINE__, "Cannot make \"$maindir/data_$partname" . '_codonpos.phy".');
					}
					print(OUT ' ' . $ntax . ' ' . scalar(@{$part2site{$partname}}) . " G\n");
					print(OUT "G 3\n");
					for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
						print(OUT ($i % 3) + 1);
						if ($i + 1 != @{$part2site{$partname}}) {
							print(OUT ' ');
						}
					}
					print(OUT "\n");
					for (my $i = 0; $i < @taxa; $i ++) {
						printf(OUT "Tax%-5d  ", $i + 1);
						foreach my $site (@{$part2site{$partname}}) {
							print(OUT $seqs{$taxa[$i]}[$site - 1]);
						}
						print(OUT "\n");
					}
					close(OUT);
				}
			}
		}
	}
	# Output NEXUS partitioning command for genes and codon positions on whole
	if (@partname == 1 && @subpart == 3) {
		unless (open(NEXUS, "> part_whole_gene_codonpos.nex")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/part_whole_gene_codonpos.nex\".");
		}
		print(NEXUS "#NEXUS\n\nBegin PAUP;\n" . &part_codonpos('PAUP') . "End;\n");
		close(NEXUS);
	}
	elsif (@partname > 1) {
		unless (open(NEXUS, "> part_whole_gene.nex")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/part_whole_gene.nex\".");
		}
		print(NEXUS "#NEXUS\n\nBegin PAUP;\n" . &whole_gene('PAUP') . "End;\n");
		close(NEXUS);
		if (@subpart > 0) {
			unless (open(NEXUS, "> part_whole_gene_codonpos.nex")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/part_whole_gene_codonpos.nex\".");
			}
			print(NEXUS "#NEXUS\n\nBegin PAUP;\n" . &whole_gene_codonpos('PAUP') . "End;\n");
			close(NEXUS);
		}
	}
	# BASEML mode
	if ($calcwhole && $lnlcalc eq 'baseml') {
		# Convert PHYLIP to PAML for genes
		if (@partname > 1) {
			unless (open(PAML, "> data_whole_gene.phy")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_whole_gene.phy\".");
			}
			print(PAML " $ntax $nchar G\n");
			print(PAML 'G ' . scalar(@partname));
			foreach my $partname (@partname) {
				print(PAML ' ' . scalar(@{$part2site{$partname}}));
			}
			print(PAML "\n");
			for (my $i = 0; $i < @taxa; $i ++) {
				printf(PAML "Tax%-5d  ", $i + 1);
				for (my $j = 0; $j < $nchar; $j ++) {
					print(PAML $seqs{$taxa[$i]}[$j]);
				}
				print(PAML "\n");
			}
			close(PAML);
		}
		# Convert PHYLIP to PAML for genes and codon positions
		if (@subpart > 0) {
			unless (open(PAML, "> data_whole_gene_codonpos.phy")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/data_whole_gene_codonpos.phy\".");
			}
			print(PAML " $ntax $nchar G\n");
			print(PAML 'G ' . (@partname - (@subpart / 3) + @subpart) . "\n");
			my $num = 1;
			my $tempnchar = 0;
			foreach my $partname (@partname) {
				if ($partname =~ /_P$/i) {
					for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
						$tempnchar ++;
						print(PAML ($num + ($i % 3)));
						if ($tempnchar != $nchar) {
							print(PAML ' ');
						}
					}
					$num += 3;
				}
				else {
					for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
						$tempnchar ++;
						print(PAML $num);
						if ($tempnchar != $nchar) {
							print(PAML ' ');
						}
					}
					$num ++;
				}
			}
			print(PAML "\n");
			for (my $i = 0; $i < @taxa; $i ++) {
				printf(PAML "Tax%-5d  ", $i + 1);
				for (my $j = 0; $j < $nchar; $j ++) {
					print(PAML $seqs{$taxa[$i]}[$j]);
				}
				print(PAML "\n");
			}
			close(PAML);
		}
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
}

sub setCandidateModels {
	if ($lnlcalc eq 'paup') {
		# PAUP mode
		delete($heterogeneity{'AGamma'});
		delete($heterogeneity{'CodonPos1Gamma'});
		delete($heterogeneity{'CodonPos3Gamma'});
		delete($heterogeneity{'CodonPos1AGamma'});
		delete($heterogeneity{'CodonPos3AGamma'});
		delete($heterogeneity{'Gene1Gamma'});
		delete($heterogeneity{'GeneNGamma'});
		delete($heterogeneity{'Gene1AGamma'});
		delete($heterogeneity{'GeneNAGamma'});
		delete($heterogeneity{'GeneCodonPos1Gamma'});
		delete($heterogeneity{'GeneCodonPosNGamma'});
		delete($heterogeneity{'GeneCodonPos1AGamma'});
		delete($heterogeneity{'GeneCodonPosNAGamma'});
	}
	elsif ($lnlcalc eq 'baseml') {
		# BASEML mode
		delete($heterogeneity{'Invariant'});
		delete($heterogeneity{'GammaInvariant'});
	}
	elsif ($lnlcalc eq 'tf') {
		delete($model{'K81'});
		delete($model{'K81uf'});
		delete($heterogeneity{'AGamma'});
		delete($heterogeneity{'CodonPos'});
		delete($heterogeneity{'CodonPos1Gamma'});
		delete($heterogeneity{'CodonPos3Gamma'});
		delete($heterogeneity{'CodonPos1AGamma'});
		delete($heterogeneity{'CodonPos3AGamma'});
		delete($heterogeneity{'Gene'});
		delete($heterogeneity{'Gene1Gamma'});
		delete($heterogeneity{'GeneNGamma'});
		delete($heterogeneity{'Gene1AGamma'});
		delete($heterogeneity{'GeneNAGamma'});
		delete($heterogeneity{'GeneCodonPos'});
		delete($heterogeneity{'GeneCodonPos1Gamma'});
		delete($heterogeneity{'GeneCodonPosNGamma'});
		delete($heterogeneity{'GeneCodonPos1AGamma'});
		delete($heterogeneity{'GeneCodonPosNAGamma'});
	}
	else {
		&errorMessage(__LINE__, 'Invalid likelihood calculation program.');
	}
	foreach my $heterogeneity (keys(%heterogeneity)) {
		if (
			($heterogeneity =~ /^Gene/ && $heterogeneity !~ /CodonPos/ && @partname < 2) || 
			($heterogeneity =~ /CodonPos/ && @subpart < 3) || 
			($heterogeneity =~ /^CodonPos/ && @partname < 2)
		) {
			delete($heterogeneity{$heterogeneity});
		}
	}
	# Candidate models of each partition
	foreach my $partname (@allpart) {
		if (($partname ne 'whole' || $calcwhole) && ($partname !~ /_P$/i || !$codonseparateonly)) {
			foreach my $model (keys(%model)) {
				$models{$partname}{$model} = 1;
			}
			foreach my $heterogeneity (keys(%heterogeneity)) {
				if (
					($partname ne 'whole' && $partname !~ /_P$/i && $heterogeneity !~ /^Gene/ && $heterogeneity !~ /^CodonPos/) || 
					($partname =~ /_P$/i && $heterogeneity ne 'AGamma' && $heterogeneity !~ /^Gene/) || 
					($partname eq 'whole' && $heterogeneity ne 'AGamma' && $heterogeneity !~ /^Gene/ && $heterogeneity !~ /^CodonPos/) || 
					($partname eq 'whole' && $heterogeneity eq 'AGamma' && @partname < 2 && @subpart < 3) || 
					($partname eq 'whole' && $heterogeneity eq 'Gene' && @partname > 1) || 
					($partname eq 'whole' && $heterogeneity =~ /Gene(?:1|N)Gamma/ && @partname > 1) || 
					($partname eq 'whole' && $heterogeneity =~ /Gene(?:1|N)AGamma/ && @partname > 1 && @subpart < 3) || 
					($partname eq 'whole' && $heterogeneity =~ /^GeneCodonPos/ && @partname > 0 && @subpart > 0)
				) {
					$heterogeneities{$partname}{$heterogeneity} = 1;
				}
			}
		}
	}
}

sub runChiSquareTest {
	print("Testing homogeneity of base composition across taxa...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	unless (mkdir('Chisq')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/Chisq\" directory.");
	}
	use Statistics::ChisqIndep;
	foreach my $partname (@allpart) {
		my @obs;
		my @sites;
		my @temptaxa;
		if ($partname =~ /_P[1-3]$/i) {
			@sites = @{$subpart{$partname}};
		}
		elsif ($partname eq 'whole') {
			@sites = (1..$nchar);
		}
		else {
			@sites = @{$part2site{$partname}};
		}
		foreach my $taxon (@taxa) {
			my @taxobs = (0, 0, 0, 0);
			foreach my $site (@sites) {
				if ($seqs{$taxon}[$site - 1] =~ /^A$/i) {
					$taxobs[0] ++;
				}
				elsif ($seqs{$taxon}[$site - 1] =~ /^C$/i) {
					$taxobs[1] ++;
				}
				elsif ($seqs{$taxon}[$site - 1] =~ /^G$/i) {
					$taxobs[2] ++;
				}
				elsif ($seqs{$taxon}[$site - 1] =~ /^(?:T|U)$/i) {
					$taxobs[3] ++;
				}
			}
			if ($taxobs[0] + $taxobs[1] + $taxobs[2] + $taxobs[3] > 0) {
				push(@obs, \@taxobs);
				push(@temptaxa, $taxon);
			}
		}
		my $chi = new Statistics::ChisqIndep;
		$chi->load_data(\@obs);
		if ($chi->{valid}) {
			unless (open(RESULT, "> Chisq/chisq_$partname.txt")) {
				&errorMessage(__LINE__, "Cannot make \"$maindir/Chisq/chisq_$partname.txt\".");
			}
			
			#output the contingency table
			my $obs = $chi->{obs}; # observed values
			my $exp = $chi->{expected}; # expected values
			my $rtotals = $chi->{rtotals}; # row totals
			my $ctotals = $chi->{ctotals}; # column totals
			my $p_value = $chi->{p_value}; # p value
			print(RESULT "Type of Nucleotides: " . $chi->{cols} . "\n");
			print(RESULT "Number of Taxa: " . $chi->{rows} . "\n"); 
			print(RESULT "Degree of Freedom: " . $chi->{df} . "\n");
			print(RESULT "Total Count: " . $chi->{total} . "\n");
			print(RESULT "Chi-square Statistic: " . $chi->{chisq_statistic} . "\n");
			print(RESULT "p-value: " . $p_value . "\n");
			if ($chi->{warning}) {
				print(RESULT "Warning: some of the cell counts might be too low.\n");
			}
			print(RESULT "\n");
			printf(RESULT "%*s ", $taxlength, ' ');
			foreach ('         A', '         C', '         G', '         T') {
				print(RESULT "\t$_");
			}
			print(RESULT "\t    rtotal\n"); 
			for (my $i = 0; $i < $chi->{rows}; $i ++) {
				printf(RESULT "%-*s \t", $taxlength, $temptaxa[$i]);
				for (my $j = 0; $j < $chi->{cols}; $j ++) {
					printf(RESULT "%10d\t", $obs->[$i]->[$j]);
				}
				printf(RESULT "%10d\n", $rtotals->[$i]);
				printf(RESULT "%*s ", $taxlength, ' ');
				for (my $j = 0; $j < $chi->{cols}; $j ++) {
					printf(RESULT "\t%10f", $exp->[$i]->[$j]);
				}
				print(RESULT "\n"); 
			}
			printf(RESULT "%-*s \t", $taxlength, 'ctotal');
			for (my $j = 0; $j < $chi->{cols}; $j ++) {
				printf(RESULT "%10d\t", $ctotals->[$j]);
			}
			#output total counts
			printf(RESULT "%10d\n", $chi->{total});
			close(RESULT);
			if ($p_value <= $siglevel && $chisq == 1) {
				delete($models{$partname}{'F81'});
				delete($models{$partname}{'HKY85'});
				delete($models{$partname}{'TN93'});
				delete($models{$partname}{'K81uf'});
				delete($models{$partname}{'J1'});
				delete($models{$partname}{'J2'});
				delete($models{$partname}{'TIM'});
				delete($models{$partname}{'TVM'});
				delete($models{$partname}{'GTR'});
			}
		}
		else {
			&errorMessage(__LINE__, 'Cannot run chi-square test.');
		}
	}
	no Statistics::ChisqIndep;
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub setModelParameters {
	print("Counting model parameters...\n");
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				# the number of branch lengths
				$nparam{$partname}{$model . '_' . $heterogeneity} = ($ntax * 2) - 3;
				# the number of parameters of nucleotide substitution rate matrix
				if ($model eq 'JC69') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 0;
				}
				elsif ($model eq 'F81') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				}
				elsif ($model eq 'K80') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				}
				elsif ($model eq 'HKY85') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 4;
				}
				elsif ($model eq 'TN93ef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				}
				elsif ($model eq 'TN93') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				}
				elsif ($model eq 'K81') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				}
				elsif ($model eq 'K81uf') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				}
				elsif ($model eq 'J1ef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				}
				elsif ($model eq 'J1') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 6;
				}
				elsif ($model eq 'J2ef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				}
				elsif ($model eq 'J2') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 6;
				}
				elsif ($model eq 'TIMef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				}
				elsif ($model eq 'TIM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 6;
				}
				elsif ($model eq 'TVMef') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 4;
				}
				elsif ($model eq 'TVM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 7;
				}
				elsif ($model eq 'SYM') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				}
				elsif ($model eq 'GTR') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 8;
				}
				# the number of parameters of site rate heterogeneity
				if ($heterogeneity eq 'Homogeneous') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 0;
				}
				elsif ($heterogeneity eq 'Gamma') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				}
				elsif ($heterogeneity eq 'Invariant') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 1;
				}
				elsif ($heterogeneity eq 'GammaInvariant') {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				}
				elsif ($heterogeneity eq 'AGamma' && (($partname ne 'whole' && $partname !~ /_P$/i) || $partname eq 'whole' && @partname < 2 && @subpart < 3)) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				}
				elsif ($heterogeneity eq 'CodonPos' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 2;
				}
				elsif ($heterogeneity eq 'CodonPos1Gamma' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 3;
				}
				elsif ($heterogeneity eq 'CodonPos3Gamma' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 5;
				}
				elsif ($heterogeneity eq 'CodonPos1AGamma' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 4;
				}
				elsif ($heterogeneity eq 'CodonPos3AGamma' && $partname =~ /_P$/i) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += 6;
				}
				elsif ($heterogeneity eq 'Gene' && $partname eq 'whole' && @partname > 1) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname - 1;
				}
				elsif ($heterogeneity eq 'Gene1Gamma' && $partname eq 'whole' && @partname > 1) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname;
				}
				elsif ($heterogeneity eq 'GeneNGamma' && $partname eq 'whole' && @partname > 1) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += (@partname * 2) - 1;
				}
				elsif ($heterogeneity eq 'Gene1AGamma' && $partname eq 'whole' && @partname > 1 && @subpart < 3) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname + 1;
				}
				elsif ($heterogeneity eq 'GeneNAGamma' && $partname eq 'whole' && @partname > 1 && @subpart < 3) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname * 2;
				}
				elsif ($heterogeneity eq 'GeneCodonPos' && $partname eq 'whole' && @partname > 0 && @subpart > 0) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname - (@subpart / 3) + @subpart - 1;
				}
				elsif ($heterogeneity eq 'GeneCodonPos1Gamma' && $partname eq 'whole' && @partname > 0 && @subpart > 0) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname - (@subpart / 3) + @subpart;
				}
				elsif ($heterogeneity eq 'GeneCodonPosNGamma' && $partname eq 'whole' && @partname > 0 && @subpart > 0) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += ((@partname - (@subpart / 3) + @subpart) * 2) - 1;
				}
				elsif ($heterogeneity eq 'GeneCodonPos1AGamma' && $partname eq 'whole' && @partname > 0 && @subpart > 0) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += @partname - (@subpart / 3) + @subpart + 1;
				}
				elsif ($heterogeneity eq 'GeneCodonPosNAGamma' && $partname eq 'whole' && @partname > 0 && @subpart > 0) {
					$nparam{$partname}{$model . '_' . $heterogeneity} += (@partname - (@subpart / 3) + @subpart) * 2;
				}
				else {
					&errorMessage(__LINE__, "Irregular model \"$model" . "_$heterogeneity\" at $partname.");
				}
			}
		}
	}
#	foreach (keys(%nparam)) {
#		print("$_ : $nparam{$_}\n");
#	}
	print("done.\n\n");
}

sub makeStartingTree {
	print("Generating tree topology for parameter optimization...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	if ($difftreeeachpart && @partname > 1) {
		my $child = 0;
		$? = 0;
		foreach my $partname ('whole', @partname) {
			if (my $pid = fork()) {
				$child ++;
				if ($child == $numthreads) {
					if (wait == -1) {
						$child = 0;
					}
					else {
						$child --;
					}
				}
				if ($?) {
					&errorMessage(__LINE__, 'Tree building error.');
				}
				next;
			}
			else {
				&starttreecmd($partname);
				exit;
			}
		}
		while (wait != -1) {
			if ($?) {
				&errorMessage(__LINE__, 'Tree building error.');
			}
		}
	}
	else {
		&starttreecmd('whole');
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub starttreecmd {
	my $partname = shift;
	if ($lnlcalc eq 'paup') {
		# PAUP mode
		unless (open(CMD, "> Logs/maketree_$partname.nex")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/maketree_$partname.nex\".");
		}
		print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n");
		if ($algorithm eq 'NJJC') {
			print(CMD "\tDSet Distance=JC BaseFreq=Equal Rates=Equal PInvar=0;\n\tNJ ShowTree=No BreakTies=Random TreeFile=starttree_$partname.tre Replace=Yes;\n");
		}
		elsif ($algorithm eq 'NJLogDet') {
			print(CMD "\tDSet Distance=LogDet BaseFreq=Equal Rates=Equal PInvar=0;\n\tNJ ShowTree=No BreakTies=Random TreeFile=starttree_$partname.tre Replace=Yes;\n");
		}
		elsif ($algorithm eq 'SARandom') {
			print(CMD "\tPSet Collapse=No;\n\tHSearch Start=Stepwise AddSeq=Random NReps=1 Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
		}
		elsif ($algorithm eq 'SAClosest') {
			print(CMD "\tPSet Collapse=No;\n\tHSearch Start=Stepwise AddSeq=Closest Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
		}
		elsif ($algorithm eq 'SAFurthest') {
			print(CMD "\tPSet Collapse=No;\n\tHSearch Start=Stepwise AddSeq=Furthest Swap=None MulTrees=No;\n\tSaveTrees Format=AltNexus File=starttree_$partname.tre Replace=Yes;\n");
		}
		else {
			&errorMessage(__LINE__, 'Cannot recognize algorithm.');
		}
		print(CMD "\tQuit WarnTSave=No;\nEnd;\n");
		close(CMD);
		if (system("$paupcmd Logs/maketree_$partname.nex > Logs/maketree_$partname.txt")) {
			&errorMessage(__LINE__, "Cannot run $paupcmd. Please check $paupcmd.");
		}
	}
	else {
		# BASEML/TF mode
		# make distance matrix by dnadist
		unless (mkdir("maketree_$partname")) {
			&errorMessage(__LINE__, "Cannot make \"maketree_$partname\" directory.");
		}
		unless (chdir("maketree_$partname")) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		unless (open(CMD, "> ../Logs/maketree_$partname.dnadist")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/maketree_$partname.dnadist\".");
		}
		print(CMD "../data_$partname.phy\nI\n2\n");
		if ($algorithm eq 'NJJC') {
			print(CMD "D\nD\nY\n");
		}
		elsif ($algorithm eq 'NJLogDet') {
			print(CMD "D\nD\nD\nY\n");
		}
		else {
			&errorMessage(__LINE__, 'Cannot recognize algorithm.');
		}
		close(CMD);
		if (system("$dnadist < ../Logs/maketree_$partname.dnadist > ../Logs/maketree_$partname" . '_dnadist.txt')) {
			&errorMessage(__LINE__, "Cannot run $dnadist. Please check $dnadist.");
		}
		unless (rename('outfile', 'infile')) {
			&errorMessage(__LINE__, "Cannot move \"$maindir/maketree_$partname/outfile\".");
		}
		# make NJ tree by neighbor
		unless (open(CMD, "> ../Logs/maketree_$partname.neighbor")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/maketree_$partname.neighbor\".");
		}
		{
			my $seed = substr(time(), -9);
			if ($seed % 2 == 0) {
				$seed ++;
			}
			print(CMD "J\n$seed\n2\n3\nY\n");
		}
		close(CMD);
		if (system("$neighbor < ../Logs/maketree_$partname.neighbor > ../Logs/maketree_$partname" . '_neighbor.txt')) {
			&errorMessage(__LINE__, "Cannot run $neighbor. Please check $neighbor.");
		}
		unless (rename('outfile', "../Logs/maketree_$partname.txt")) {
			&errorMessage(__LINE__, "Cannot move \"$maindir/outfile\".");
		}
		unless (rename('infile', "../Logs/$partname.dist")) {
			&errorMessage(__LINE__, "Cannot move \"$maindir/infile\".");
		}
		unless (open(IN, '< outtree')) {
			&errorMessage(__LINE__, "Cannot open \"$maindir/outtree\".");
		}
		unless (open(OUT, "> ../starttree_$partname.nwk")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/starttree_$partname.nwk\".");
		}
		while (<IN>) {
			s/-\d+(?:\.\d+)?/0/g;
			s/(.+)\r?\n?/$1/;
			s/;$/;\n/;
			print(OUT);
		}
		close(OUT);
		close(IN);
		unlink('outtree');
		unless (chdir('..')) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		unless (rmdir("maketree_$partname")) {
			&errorMessage(__LINE__, 'Cannot remove temporary directory.');
		}
	}
}

sub setSampleSizes {
	print("Counting sample sizes...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	# sample size 0 : pscore
	# sample size 1 : sum of (the number of character states - 1)
	# sample size 2 : sum of the number of character states
	# sample size 3 : the number of sites = alignment length
	# sample size 4 : the number of variable characters
	# sample size 5 : (sample size 3) x the number of OTUs
	my $child = 0;
	$? = 0;
	foreach my $partname (@allpart) {
		if (my $pid = fork()) {
			$child ++;
			if ($child == $numthreads) {
				if (wait == -1) {
					$child = 0;
				}
				else {
					$child --;
				}
			}
			if ($?) {
				&errorMessage(__LINE__, 'Sample size counting error.');
			}
			next;
		}
		else {
			&calcpscore($partname);
			exit;
		}
	}
	while (wait != -1) {
		if ($?) {
			&errorMessage(__LINE__, 'Sample size counting error.');
		}
	}
#	foreach my $partname (@allpart) {
#		&calcpscore($partname);
#	}
	foreach my $partname (@allpart) {
		unless (open(SCORE, "< Logs/pscore_$partname.txt")) {
			&errorMessage(__LINE__, "Cannot open \"$maindir/Logs/pscore_$partname.txt\".");
		}
		while (<SCORE>) {
			# PAUP mode
			if (/^1\t(\d+)/ && $lnlcalc eq 'paup') {
				$samplesize{$partname}[0] = $1;
				last;
			}
			# BASEML/TF mode
			elsif (/^requires a total of\s*(\d+)/ && $lnlcalc ne 'paup') {
				$samplesize{$partname}[0] = $1;
				last;
			}
		}
		close(SCORE);
	}
	my @numstate;
	for (my $i = 0; $i < $nchar; $i ++) {
		my %state;
		foreach my $taxon (@taxa) {
			if ($seqs{$taxon}[$i] =~ /^[ACGTMRWSYKVHDBN]$/i) {
				$state{$seqs{$taxon}[$i]} = 1;
			}
			elsif ($seqs{$taxon}[$i] =~ /^U$/i) {
				$state{'T'} = 1;
			}
			elsif ($seqs{$taxon}[$i] =~ /^-$/) {
				$state{'N'} = 1;
			}
			elsif ($seqs{$taxon}[$i] =~ /^\?$/) {
				$state{'N'} = 1;
			}
		}
		if (exists($state{'M'}) && (exists($state{'A'}) || exists($state{'C'}))) {
			delete($state{'M'});
		}
		if (exists($state{'R'}) && (exists($state{'A'}) || exists($state{'G'}))) {
			delete($state{'R'});
		}
		if (exists($state{'W'}) && (exists($state{'A'}) || exists($state{'T'}))) {
			delete($state{'W'});
		}
		if (exists($state{'S'}) && (exists($state{'C'}) || exists($state{'G'}))) {
			delete($state{'S'});
		}
		if (exists($state{'Y'}) && (exists($state{'C'}) || exists($state{'T'}))) {
			delete($state{'Y'});
		}
		if (exists($state{'K'}) && (exists($state{'G'}) || exists($state{'T'}))) {
			delete($state{'K'});
		}
		if (exists($state{'V'}) && (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'M'}) || exists($state{'R'}) || exists($state{'S'}))) {
			delete($state{'V'});
		}
		if (exists($state{'H'}) && (exists($state{'A'}) || exists($state{'C'}) || exists($state{'T'}) || exists($state{'M'}) || exists($state{'W'}) || exists($state{'Y'}))) {
			delete($state{'H'});
		}
		if (exists($state{'D'}) && (exists($state{'A'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'R'}) || exists($state{'W'}) || exists($state{'K'}))) {
			delete($state{'D'});
		}
		if (exists($state{'B'}) && (exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'S'}) || exists($state{'Y'}) || exists($state{'K'}))) {
			delete($state{'B'});
		}
		if (exists($state{'V'}) && exists($state{'H'}) && exists($state{'D'}) && exists($state{'B'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'V'}) && exists($state{'H'}) && exists($state{'D'}) || exists($state{'V'}) && exists($state{'H'}) && exists($state{'B'}) || exists($state{'V'}) && exists($state{'D'}) && exists($state{'B'}) || exists($state{'H'}) && exists($state{'D'}) && exists($state{'B'})) {
			$numstate[$i] = 1;
		}
		elsif (exists($state{'V'}) && exists($state{'H'})) {
			if (exists($state{'K'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'V'}) && exists($state{'D'})) {
			if (exists($state{'Y'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'V'}) && exists($state{'B'})) {
			if (exists($state{'W'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'H'}) && exists($state{'D'})) {
			if (exists($state{'S'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'H'}) && exists($state{'B'})) {
			if (exists($state{'R'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'D'}) && exists($state{'B'})) {
			if (exists($state{'M'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'V'})) {
			if (exists($state{'T'}) || exists($state{'W'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'K'}) || exists($state{'Y'}) && exists($state{'K'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'H'})) {
			if (exists($state{'G'}) || exists($state{'R'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'K'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'D'})) {
			if (exists($state{'C'}) || exists($state{'M'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'Y'}) || exists($state{'S'}) && exists($state{'Y'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'B'})) {
			if (exists($state{'A'}) || exists($state{'M'}) && exists($state{'R'}) || exists($state{'M'}) && exists($state{'W'}) || exists($state{'R'}) && exists($state{'W'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 3;
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'Y'}) && exists($state{'K'})) {
			if (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 3;
			}
			else {
				$numstate[$i] = 2;
			}
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'W'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'Y'}) && exists($state{'K'})) {
			$numstate[$i] = 1;
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) && exists($state{'Y'}) || exists($state{'M'}) && exists($state{'R'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'W'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'S'}) && exists($state{'K'}) || exists($state{'M'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'W'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'Y'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'S'}) && exists($state{'K'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'M'}) && exists($state{'K'}) || exists($state{'R'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'S'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'M'}) && exists($state{'R'}) || exists($state{'M'}) && exists($state{'W'}) || exists($state{'M'}) && exists($state{'S'}) || exists($state{'M'}) && exists($state{'Y'}) || exists($state{'R'}) && exists($state{'W'}) || exists($state{'R'}) && exists($state{'S'}) || exists($state{'R'}) && exists($state{'K'}) || exists($state{'W'}) && exists($state{'Y'}) || exists($state{'W'}) && exists($state{'K'}) || exists($state{'S'}) && exists($state{'Y'}) || exists($state{'S'}) && exists($state{'K'}) || exists($state{'Y'}) && exists($state{'K'})) {
			if (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'M'}) || exists($state{'R'}) || exists($state{'W'}) || exists($state{'S'}) || exists($state{'Y'}) || exists($state{'K'})) {
			if (exists($state{'A'}) && exists($state{'C'}) || exists($state{'A'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) || exists($state{'C'}) && exists($state{'T'}) || exists($state{'G'}) && exists($state{'T'})) {
				$numstate[$i] = 3;
			}
			elsif (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'})) {
				$numstate[$i] = 2;
			}
			else {
				$numstate[$i] = 1;
			}
		}
		elsif (exists($state{'A'}) && exists($state{'C'}) && exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 4;
		}
		elsif (exists($state{'A'}) && exists($state{'C'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'C'}) && exists($state{'T'}) || exists($state{'A'}) && exists($state{'G'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 3;
		}
		elsif (exists($state{'A'}) && exists($state{'C'}) || exists($state{'A'}) && exists($state{'G'}) || exists($state{'A'}) && exists($state{'T'}) || exists($state{'C'}) && exists($state{'G'}) || exists($state{'C'}) && exists($state{'T'}) || exists($state{'G'}) && exists($state{'T'})) {
			$numstate[$i] = 2;
		}
		elsif (exists($state{'A'}) || exists($state{'C'}) || exists($state{'G'}) || exists($state{'T'}) || exists($state{'N'})) {
			$numstate[$i] = 1;
		}
		$samplesize{'whole'}[2] += $numstate[$i];
		unless ($numstate[$i]) {
			&errorMessage(__LINE__, 'The number of state of ' . ($i + 1) . 'th column is unable to count.');
		}
		elsif ($numstate[$i] > 1) {
			$samplesize{'whole'}[4] ++;
		}
	}
	undef(%seqs);
	$samplesize{'whole'}[3] = $nchar;
	$samplesize{'whole'}[1] = $samplesize{'whole'}[2] - $samplesize{'whole'}[3];
	$samplesize{'whole'}[5] = $nchar * $ntax;
	unless (defined($samplesize{'whole'}[4])) {
		$samplesize{'whole'}[4] = 0;
	}
	if (@partname > 1) {
		foreach my $partname (@partname) {
			foreach my $site (@{$part2site{$partname}}) {
				$samplesize{$partname}[2] += $numstate[$site - 1];
				if ($numstate[$site - 1] > 1) {
					$samplesize{$partname}[4] ++;
				}
			}
			$samplesize{$partname}[3] = scalar(@{$part2site{$partname}});
			$samplesize{$partname}[1] = $samplesize{$partname}[2] - $samplesize{$partname}[3];
			$samplesize{$partname}[5] = scalar(@{$part2site{$partname}}) * $ntax;
			unless (defined($samplesize{$partname}[4])) {
				$samplesize{$partname}[4] = 0;
			}
		}
	}
	if ($dividecodon) {
		foreach my $partname (@subpart) {
			foreach my $site (@{$subpart{$partname}}) {
				$samplesize{$partname}[2] += $numstate[$site - 1];
				if ($numstate[$site - 1] > 1) {
					$samplesize{$partname}[4] ++;
				}
			}
			$samplesize{$partname}[3] = scalar(@{$subpart{$partname}});
			$samplesize{$partname}[1] = $samplesize{$partname}[2] - $samplesize{$partname}[3];
			$samplesize{$partname}[5] = scalar(@{$subpart{$partname}}) * $ntax;
			unless (defined($samplesize{$partname}[4])) {
				$samplesize{$partname}[4] = 0;
			}
		}
	}
	foreach my $partname (@allpart) {
#		print("$partname : @{$samplesize{$partname}}\n");
		foreach (0 .. 5) {
			unless (defined($samplesize{$partname}[$_])) {
				&errorMessage(__LINE__, 'Sample size counting error.');
			}
		}
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcpscore {
	my $partname = shift;
	if ($lnlcalc eq 'paup') {
		# PAUP mode
		unless (open(CMD, "> Logs/pscore_$partname.nex")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/pscore_$partname.nex\".");
		}
		print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n\tGetTrees File=starttree_");
		if ($difftreeeachpart) {
			if ($partname =~ /^(.+_P)[1-3]$/i) {
				print(CMD $1);
			}
			else {
				print(CMD $partname);
			}
		}
		else {
			print(CMD 'whole');
		}
		print(CMD ".tre From=1 To=1 Unrooted=Yes;\n");
		print(CMD "\tPScores 1 / ScoreFile=Logs/pscore_$partname.txt Replace=Yes;\n");
		print(CMD "\tQuit WarnTSave=No;\nEnd;\n");
		close(CMD);
		if (system("$paupcmd Logs/pscore_$partname.nex > Logs/pscore_$partname" . '_paup.txt')) {
			&errorMessage(__LINE__, "Cannot run $paupcmd. Please check $paupcmd.");
		}
	}
	else {
		# BASEML/TF mode
		unless (open(CMD, "> Logs/pscore_$partname.dnapars")) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/pscore_$partname.dnapars\".");
		}
		print(CMD "../data_$partname.phy\nU\nI\n2\n6\nY\n../starttree_");
		if ($difftreeeachpart) {
			if ($partname =~ /^(.+_P)[1-3]$/i) {
				print(CMD $1);
			}
			else {
				print(CMD $partname);
			}
		}
		else {
			print(CMD 'whole');
		}
		print(CMD ".nwk\n");
		close(CMD);
		unless (mkdir("pscore_$partname")) {
			&errorMessage(__LINE__, "Cannot make \"pscore_$partname\" directory.");
		}
		unless (chdir("pscore_$partname")) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		if (system("$dnapars < ../Logs/pscore_$partname.dnapars > ../Logs/pscore_$partname" . '_dnapars.txt')) {
			&errorMessage(__LINE__, "Cannot run $dnapars. Please check $dnapars.");
		}
		unless (rename('outfile', "../Logs/pscore_$partname.txt")) {
			&errorMessage(__LINE__, "Cannot move \"$maindir/pscore_$partname/outfile\".");
		}
		unless (chdir('..')) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		unless (rmdir("pscore_$partname")) {
			&errorMessage(__LINE__, 'Cannot remove temporary directory.');
		}
	}
}

sub analyzeEachPartition {
	print("Analyzing each partition under the models...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	unless (mkdir('Scores')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/Scores\" directory.");
	}
	my $child = 0;
	$? = 0;
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				if (($partname ne 'whole' || $calcwhole) && ($partname !~ /_P$/i || !$codonseparateonly)) {
					if (my $pid = fork()) {
						$child ++;
						if ($child == $numthreads) {
							if (wait == -1) {
								$child = 0;
							}
							else {
								$child --;
							}
						}
						if ($?) {
							&errorMessage(__LINE__, 'Likelihood calculation error.');
						}
						next;
					}
					else {
						&calcL($partname, $model, $heterogeneity);
						exit;
					}
				}
			}
		}
	}
	while (wait != -1) {
		if ($?) {
			&errorMessage(__LINE__, 'Likelihood calculation error.');
		}
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcL {
	my ($partname, $model, $heterogeneity) = @_;
	print("Calculating likelihood under $model" . '_' . "$heterogeneity model on $partname sequences...\n");
	if ($lnlcalc eq 'paup') {
		# PAUP mode
		unless (open(CMD, '> ' . $partname . '_' . $model . '_' . $heterogeneity . '.nex')) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/$partname" . '_' . $model . '_' . "$heterogeneity.nex\".");
		}
		print(CMD '#NEXUS' . "\n\nBegin PAUP;\n\tSet ErrorBeep=No QueryBeep=No KeyBeep=No NotifyBeep=No ErrorStop=No WarnReset=No WarnTree=No WarnTSave=No WarnBlkName=No WarnRoot=No WarnRedef=No;\n\tExecute data_$partname.nex;\n\tDefaults LScores LongFmt=Yes;\n\tSet Criterion=Likelihood;\n\tGetTrees File=starttree_");
		if ($difftreeeachpart) {
			if ($partname =~ /^(.+_P)[1-3]$/i) {
				print(CMD $1);
			}
			else {
				print(CMD $partname);
			}
		}
		else {
			print(CMD 'whole');
		}
		print(CMD ".tre From=1 To=1 Unrooted=Yes;\n");
		if ($heterogeneity =~ /GeneCodonPos/) {
			print(CMD "\tExecute part_whole_gene_codonpos.nex;\n");
		}
		elsif ($heterogeneity =~ /CodonPos/) {
			print(CMD "\tExecute part_$partname" . "_codonpos.nex;\n");
		}
		elsif ($heterogeneity =~ /Gene/) {
			print(CMD "\tExecute part_whole_gene.nex;\n");
		}
		print(CMD "\t" . &makePAUPLSet($model, $heterogeneity) . "\n");
		print(CMD "\tLScores 1 / ScoreFile=Scores/" . $partname . '_' . $model . '_' . $heterogeneity . ".txt Replace=Yes;\n");
		print(CMD "\tQuit WarnTSave=No;\n");
		print(CMD 'End;' . "\n");
		close(CMD);
		if (system("$paupcmd " . $partname . '_' . $model . '_' . $heterogeneity . '.nex > Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.txt')) {
			&errorMessage(__LINE__, "Cannot run $paupcmd. Please check $paupcmd.");
		}
		unless (rename($partname . '_' . $model . '_' . $heterogeneity . '.nex', 'Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.nex')) {
			&errorMessage(__LINE__, "Cannot move \"$maindir/" . $partname . '_' . $model . '_' . $heterogeneity . ".nex\".");
		}
	}
	elsif ($lnlcalc eq 'tf') {
		# TF mode
		unless (open(CMD, '> Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.tl')) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/$partname" . '_' . $model . '_' . "$heterogeneity.tl\".");
		}
		my $tempmodel = &makeTFModel($model . '_' . $heterogeneity);
		print(CMD "report:=ReconstructPhylogeny[\n\t\"data_$partname.phy\",\n\tSubstitutionModel->$tempmodel,\n");
		print(CMD "\tTree->\"starttree_");
		if ($difftreeeachpart) {
			if ($partname =~ /^(.+_P)[1-3]$/i) {
				print(CMD $1);
			}
			else {
				print(CMD $partname);
			}
		}
		else {
			print(CMD 'whole');
		}
		print(CMD ".nwk\",\n");
		print(CMD "\tWithEdgeSupport->False,\n\tAcceptFlatness->True,\n\tVerbose->False\n],\nOprec[\n\t20,\n\tSave[\n\t\treport|1|Likelihood,\n\t\t\"Scores/$partname\_$model\_$heterogeneity.txt\"\n\t]\n]\n");
		close(CMD);
		if (system("$tf Logs/" . $partname . '_' . $model . '_' . $heterogeneity . '.tl')) {
			&errorMessage(__LINE__, "Cannot run $tf. Please check $tf.");
		}
	}
	elsif ($lnlcalc eq 'baseml') {
		# BASEML mode
		unless (open(CMD, '> Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.paml')) {
			&errorMessage(__LINE__, "Cannot make \"$maindir/Logs/$partname" . '_' . $model . '_' . "$heterogeneity.paml\".");
		}
		if ($heterogeneity =~ /^GeneCodonPos/) {
			print(CMD "seqfile = ../data_whole_gene_codonpos.phy\n");
		}
		elsif ($heterogeneity =~ /^Gene/) {
			print(CMD "seqfile = ../data_whole_gene.phy\n");
		}
		elsif ($heterogeneity =~ /^CodonPos/) {
			print(CMD "seqfile = ../data_$partname" . "_codonpos.phy\n");
		}
		else {
			print(CMD "seqfile = ../data_$partname.phy\n");
		}
		print(CMD "treefile = ../starttree_");
		if ($difftreeeachpart) {
			if ($partname =~ /^(.+_P)[1-3]$/i) {
				print(CMD $1);
			}
			else {
				print(CMD $partname);
			}
		}
		else {
			print(CMD 'whole');
		}
		print(CMD ".nwk\n");
		print(CMD "outfile = ../Scores/" . $partname . '_' . $model . '_' . $heterogeneity . ".txt\n");
		print(CMD "noisy = 0\n");
		print(CMD "verbose = 0\n");
		print(CMD "Mgene = 0\n");
		print(CMD "clock = 0\n");
		print(CMD "getSE = 0\n");
		print(CMD "RateAncestor = 0\n");
		if ($heterogeneity =~ /AGamma/) {
			print(CMD "method = 0\n");
		}
		else {
			print(CMD "method = 1\n");
		}
		print(CMD "Small_Diff = 1e-6\n");
		print(CMD "fix_blength = 0\n");
		if ($model eq 'JC69') {
			print(CMD "model = 0\n");
			print(CMD "fix_kappa = 1\n");
		}
		elsif ($model eq 'F81') {
			print(CMD "model = 2\n");
			print(CMD "fix_kappa = 1\n");
		}
		elsif ($model eq 'K80') {
			print(CMD "model = 1\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model eq 'HKY85') {
			print(CMD "model = 4\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^TN93/) {
			print(CMD "model = 6\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^K81/) {
			print(CMD "model = 9 [2 (AG CT) (AT CG)]\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^J1/) {
			print(CMD "model = 9 [3 (AG) (AC CG) (CT)]\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^J2/) {
			print(CMD "model = 9 [3 (AG) (AC AT) (CT)]\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^TIM/) {
			print(CMD "model = 9 [3 (AG) (AT CG) (CT)]\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model =~ /^TVM/) {
			print(CMD "model = 9 [4 (AC) (AG CT) (AT) (CG)]\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model eq 'SYM') {
			print(CMD "model = 7\n");
			print(CMD "fix_kappa = 0\n");
		}
		elsif ($model eq 'GTR') {
			print(CMD "model = 7\n");
			print(CMD "fix_kappa = 0\n");
		}
		if ($model =~ /^(?:JC69|K80|TN93ef|K81|J1ef|J2ef|TIMef|TVMef|SYM)$/) {
			print(CMD "nhomo = -1\n");
		}
		elsif ($basefreq eq 'Estimate') {
			print(CMD "nhomo = 1\n");
		}
		elsif ($basefreq eq 'Empirical') {
			print(CMD "nhomo = 0\n");
		}
		if ($heterogeneity eq 'Homogeneous' || $heterogeneity eq 'CodonPos' || $heterogeneity eq 'Gene' || $heterogeneity eq 'GeneCodonPos') {
			print(CMD "fix_alpha = 1\nalpha = 0\nMalpha = 0\nfix_rho = 1\nrho = 0\nnparK = 0\n");
		}
		elsif ($heterogeneity eq 'Gamma' || $heterogeneity eq 'CodonPos1Gamma' || $heterogeneity eq 'Gene1Gamma' || $heterogeneity eq 'GeneCodonPos1Gamma') {
			print(CMD "fix_alpha = 0\nalpha = 2\nMalpha = 0\nfix_rho = 1\nrho = 0\nnparK = 0\nncatG = $ngammacat\n");
		}
		elsif ($heterogeneity eq 'AGamma' || $heterogeneity eq 'CodonPos1AGamma' || $heterogeneity eq 'Gene1AGamma' || $heterogeneity eq 'GeneCodonPos1AGamma') {
			print(CMD "fix_alpha = 0\nalpha = 2\nMalpha = 0\nfix_rho = 0\nrho = 0\nnparK = 0\nncatG = $ngammacat\n");
		}
		elsif ($heterogeneity eq 'CodonPos3Gamma' || $heterogeneity eq 'GeneNGamma' || $heterogeneity eq 'GeneCodonPosNGamma') {
			print(CMD "fix_alpha = 0\nalpha = 2\nMalpha = 1\nfix_rho = 1\nrho = 0\nnparK = 0\nncatG = $ngammacat\n");
		}
		elsif ($heterogeneity eq 'CodonPos3AGamma' || $heterogeneity eq 'GeneNAGamma' || $heterogeneity eq 'GeneCodonPosNAGamma') {
			print(CMD "fix_alpha = 0\nalpha = 2\nMalpha = 1\nfix_rho = 0\nrho = 0\nnparK = 0\nncatG = $ngammacat\n");
		}
		close(CMD);
		unless (mkdir($partname . '_' . $model . '_' . $heterogeneity)) {
			&errorMessage(__LINE__, "Cannot make \"$partname" . '_' . $model . '_' . "$heterogeneity\" directory.");
		}
		unless (chdir($partname . '_' . $model . '_' . $heterogeneity)) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		if (system("$baseml ../Logs/" . $partname . '_' . $model . '_' . $heterogeneity . '.paml > ../Logs/' . $partname . '_' . $model . '_' . $heterogeneity . '.txt')) {
			&errorMessage(__LINE__, "Cannot run $baseml. Please check $baseml.");
		}
		unlink('2base.t', 'lnf', 'rates', 'rst', 'rst1', 'rub', 'in.basemlg');
		unless (chdir('..')) {
			&errorMessage(__LINE__, 'Cannot change working directory.');
		}
		unless (rmdir($partname . '_' . $model . '_' . $heterogeneity)) {
			&errorMessage(__LINE__, 'Cannot remove temporary directory.');
		}
	}
}

sub calculateCriteria {
	print("Calculating AIC, AICc, and BIC...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	foreach my $partname (@allpart) {
		foreach my $model (keys(%{$models{$partname}})) {
			foreach my $heterogeneity (keys(%{$heterogeneities{$partname}})) {
				if (open(LNL, '< Scores/' . $partname . '_' . $model . '_' . $heterogeneity . '.txt')) {
					while (<LNL>) {
						if (($lnlcalc eq 'paup' && /^1\t(\d+\.\d+)/) || ($lnlcalc eq 'baseml' && /^lnL\([^\(\)]+\):.*?-(\d+\.\d+)/) || ($lnlcalc eq 'tf' && /-(\d+\.\d+)/)) {
							$LnL{$partname}{$model . '_' . $heterogeneity} = $1;
							last;
						}
					}
					close(LNL);
					$criteria{'AIC'}{$partname}{$model . '_' . $heterogeneity} = &calcAIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity});
					if ($samplesize{$partname}[0] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc1'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[0]);
					}
					if ($samplesize{$partname}[1] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc2'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[1]);
					}
					if ($samplesize{$partname}[2] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc3'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[2]);
					}
					if ($samplesize{$partname}[3] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc4'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[3]);
					}
					if ($samplesize{$partname}[4] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc5'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[4]);
					}
					if ($samplesize{$partname}[5] - $nparam{$partname}{$model . '_' . $heterogeneity} - 1 > 0) {
						$criteria{'AICc6'}{$partname}{$model . '_' . $heterogeneity} = &calcAICc($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[5]);
					}
					$criteria{'BIC1'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[0]);
					$criteria{'BIC2'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[1]);
					$criteria{'BIC3'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[2]);
					$criteria{'BIC4'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[3]);
					$criteria{'BIC5'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[4]);
					$criteria{'BIC6'}{$partname}{$model . '_' . $heterogeneity} = &calcBIC($LnL{$partname}{$model . '_' . $heterogeneity}, $nparam{$partname}{$model . '_' . $heterogeneity}, $samplesize{$partname}[5]);
				}
			}
		}
	}
	&calcweight();
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub calcAIC {
	my $AIC = 2 * $_[0] + 2 * $_[1];
	return($AIC);
}

sub calcAICc {
	my $AICc;
	$AICc = 2 * $_[0] + 2 * $_[1] * $_[2] / ($_[2] - $_[1] - 1);
	return($AICc);
}

sub calcBIC {
	my $BIC = 2 * $_[0] + $_[1] * log($_[2]);
	return($BIC);
}

sub calcweight {
	foreach my $partname (@allpart) {
		foreach my $criterion (keys(%criteria)) {
			my $num = 0;
			my $minvalue;
			my $sum = 0;
			foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
				if ($num == 0) {
					$minvalue = $criteria{$criterion}{$partname}{$model};
					$sum = 1;
				}
				else {
					$sum += exp(($criteria{$criterion}{$partname}{$model} - $minvalue) / (-2));
				}
				$num ++;
			}
			foreach my $model (keys(%{$criteria{$criterion}{$partname}})) {
				$weight{$criterion}{$partname}{$model} = exp(($criteria{$criterion}{$partname}{$model} - $minvalue) / (-2)) / $sum;
			}
		}
	}
}

sub outputFiles {
	print("Outputting results...\n");
	unless (chdir($maindir)) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	# output results
	&outputResults();
	# output for MrBayes
	if ($output{'mrbayes'} || $output{'all'}) {
		&outforMrBayes();
	}
	# output for PAUP*
	if ($output{'paup'} || $output{'all'}) {
		&outforPAUP();
	}
	# output for Treefinder
	if ($output{'treefinder'} || $output{'all'}) {
		&outforTreefinder();
	}
	unless (chdir('..')) {
		&errorMessage(__LINE__, 'Cannot change working directory.');
	}
	print("done.\n\n");
}

sub outputResults {
	unless (mkdir('Results')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/Results\" directory.");
	}
	foreach my $partname (@allpart) {
		foreach my $criterion (keys(%criteria)) {
			if ($criteria{$criterion}{$partname}{'JC69_Homogeneous'}) {
				unless (open(OUT, "> Results/$partname" . "_$criterion.txt")) {
					&errorMessage(__LINE__, "Cannot make \"$maindir/Resilts/$partname" . '_' . "$criterion.txt\".");
				}
				print(OUT 'model                     ' . "\t");
				printf(OUT "% 19s \t", $criterion);
				print(OUT 'weight ' . "\t");
				print(OUT '               -LnL ' . "\t" . 'nparam');
				if ($criterion =~ /(\d)$/) {
					print(OUT "\t" . 'sample size' );
				}
				print(OUT "\n");
				foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
					printf(OUT "%-26s\t", $model);
					printf(OUT "% 20.12e\t", $criteria{$criterion}{$partname}{$model});
					printf(OUT "%.5f\t", $weight{$criterion}{$partname}{$model});
					printf(OUT "% 20.12e\t", $LnL{$partname}{$model});
					printf(OUT "% 6d", $nparam{$partname}{$model});
					if ($criterion =~ /(\d)$/) {
						printf(OUT "\t% 11d", $samplesize{$partname}[$1 - 1]);
					}
					print(OUT "\n");
				}
				close(OUT);
			}
		}
	}
}

sub outforPAUP {
	unless (mkdir('PAUP')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/PAUP\" directory.");
	}
	my @temppartname;
	if (@partname > 1) {
		@temppartname = @partname;
	}
	if ($dividecodon) {
		push(@temppartname, @subpart);
	}
	foreach my $criterion (keys(%criteria)) {
		foreach my $partname ('whole', @temppartname) {
			if ($criteria{$criterion}{$partname}{'JC69_Homogeneous'}) {
				&makePAUPNexus($criterion, $partname);
			}
		}
	}
}

sub makePAUPNexus {
	my ($criterion, $partname) = @_;
	my %modelPAUP;
	foreach my $model (@modelforpaup) {
		foreach my $heterogeneity (@heterogeneityforpaup) {
			$modelPAUP{$model . '_' . $heterogeneity} = 1;
		}
	}
	unless (open(OUT, '> PAUP/' . $partname . '_' . $criterion . ".nex")) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/PAUP/$partname" . '_' . "$criterion.nex\".");
	}
	print(OUT &dataNexus($partname));
	print(OUT "\nBegin PAUP;\n");
	print(OUT "\tSet Criterion=Likelihood;\n");
	my $notbest;
	foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
		print(OUT '[ ');
		printf(OUT "%-26s ", $model);
		printf(OUT "% 12.5e", $criteria{$criterion}{$partname}{$model});
		print(OUT ' ]  ');
		if ($modelPAUP{$model}) {
			unless ($notbest) {
				print(OUT '  ');
			}
			else {
				print(OUT '[ ');
			}
			my $tempcmd;
			if ($model =~ /GeneCodonPos/) {
				$tempcmd = &whole_gene_codonpos('PAUP');
			}
			elsif ($model =~ /CodonPos/) {
				$tempcmd = &part_codonpos('PAUP');
			}
			elsif ($model =~ /Gene/) {
				$tempcmd = &whole_gene('PAUP');
			}
			$tempcmd =~ s/\t//g;
			$tempcmd =~ s/\n/ /g;
			print(OUT $tempcmd);
			print(OUT &makePAUPLSet(split(/_/, $model)));
			unless ($notbest) {
				$notbest = 1;
			}
			else {
				print(OUT ' ]');
			}
		}
		print(OUT "\n");
	}
	print(OUT "End;\n");
	close(OUT);
}

sub makePAUPLSet {
	my ($model, $heterogeneity) = @_;
	my $lset1;
	my $lset2;
	if ($model eq 'JC69') {
		$lset1 = 'Nst=1 BaseFreq=Equal';
	}
	elsif ($model eq 'F81') {
		$lset1 = 'Nst=1 BaseFreq=' . $basefreq;
	}
	elsif ($model eq 'K80') {
		$lset1 = 'Nst=2 BaseFreq=Equal TRatio=Estimate';
	}
	elsif ($model eq 'HKY85') {
		$lset1 = 'Nst=2 BaseFreq=' . $basefreq . ' TRatio=Estimate';
	}
	elsif ($model eq 'TN93ef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b a a e a)';
	}
	elsif ($model eq 'TN93') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b a a e a)';
	}
	elsif ($model eq 'K81uf') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c c b a)';
	}
	elsif ($model eq 'K81') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c c b a)';
	}
	elsif ($model eq 'J1ef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c a e c)';
	}
	elsif ($model eq 'J1') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c a e c)';
	}
	elsif ($model eq 'J2ef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b a d e d)';
	}
	elsif ($model eq 'J2') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b a d e d)';
	}
	elsif ($model eq 'TIMef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c c e a)';
	}
	elsif ($model eq 'TIM') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c c e a)';
	}
	elsif ($model eq 'TVMef') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c d b e)';
	}
	elsif ($model eq 'TVM') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c d b e)';
	}
	elsif ($model eq 'SYM') {
		$lset1 = 'Nst=6 BaseFreq=Equal RMatrix=Estimate RClass=(a b c d e f)';
	}
	elsif ($model eq 'GTR') {
		$lset1 = 'Nst=6 BaseFreq=' . $basefreq . ' RMatrix=Estimate RClass=(a b c d e f)';
	}
	if ($heterogeneity eq 'Homogeneous') {
		$lset2 = ' Rates=Equal PInvar=0';
	}
	elsif ($heterogeneity eq 'Gamma') {
		$lset2 = ' Rates=Gamma Shape=Estimate ' . "NCat=$ngammacat";
	}
	elsif ($heterogeneity eq 'Invariant') {
		$lset2 = ' Rates=Equal PInvar=Estimate';
	}
	elsif ($heterogeneity eq 'GammaInvariant') {
		$lset2 = ' Rates=Gamma Shape=Estimate PInvar=Estimate ' . "NCat=$ngammacat";
	}
	elsif ($heterogeneity eq 'Gene' || $heterogeneity eq 'CodonPos' || $heterogeneity eq 'GeneCodonPos') {
		$lset2 = ' Rates=SiteSpec SiteRates=Partition:Kakusan';
	}
	my $cmd = "LSet $lset1$lset2;";
	return($cmd);
}

sub outforMrBayes {
	unless (mkdir('MrBayes')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/MrBayes\" directory.");
	}
	my @temppartname;
	if (@partname > 1) {
		@temppartname = @partname;
	}
	foreach my $criterion (keys(%criteria)) {
		foreach my $partname ('whole', @temppartname) {
			if ($partname eq 'whole' || $criteria{$criterion}{$partname}{'JC69_Homogeneous'} || $partname =~ /_P$/i && $codonseparateonly && $criteria{$criterion}{$partname . 1}{'JC69_Homogeneous'} && $criteria{$criterion}{$partname . 2}{'JC69_Homogeneous'} && $criteria{$criterion}{$partname . 3}{'JC69_Homogeneous'}) {
				&makeMrBayesNexus($criterion, $partname);
			}
		}
	}
}

sub makeMrBayesNexus {
	my ($criterion, $partname) = @_;
	my $outfile;
	my $partcriterion = 1;
	my $subpartcriterion = 1;
	if ($partname eq 'whole') {
		foreach my $temppartname (@partname) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_Homogeneous'}) {
				$partcriterion = 0;
			}
		}
		foreach my $temppartname (@subpart) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_Homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	}
	elsif ($partname =~ /_P$/i) {
		foreach my $subpartno (1,2,3) {
			unless ($criteria{$criterion}{$partname . $subpartno}{'JC69_Homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	}
	# single model
	if ($partname ne 'whole' && $partname !~ /_P$/i || $criteria{$criterion}{$partname}{'JC69_Homogeneous'}) {
		if ($partname eq 'whole' && @partname == 1 && @subpart > 0 || $partname =~ /_P$/i) {
			$outfile = $partname . '_' . $criterion . '_codonshared.nex';
		}
		else {
			$outfile = $partname . '_' . $criterion . '_single.nex';
		}
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		print(OUT &makeMrBayesCommand($criterion, 0, $partname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# proportional model (codon position singlized)
	if ($partname eq 'whole' && @partname > 1 && $partcriterion == 1) {
		if (@subpart > 0) {
			$outfile = $partname . '_' . $criterion . '_proportional_codonshared.nex';
		}
		else {
			$outfile = $partname . '_' . $criterion . '_proportional.nex';
		}
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			my $codonpos;
			if ($outfile =~ /codonshared.nex$/) {
				my %modelMrBayes;
				foreach my $tempmodel (@modelformrbayes) {
					foreach my $tempheterogeneity (@heterogeneityformrbayes) {
						$modelMrBayes{$tempmodel . '_' . $tempheterogeneity} = 1;
					}
				}
				foreach my $tempmodel (sort({$criteria{$criterion}{$temppartname}{$a} <=> $criteria{$criterion}{$temppartname}{$b}} keys(%{$criteria{$criterion}{$temppartname}}))) {
					if ($modelMrBayes{$tempmodel}) {
						if ($tempmodel =~ /CodonPos/) {
							$codonpos = 1;
						}
						last;
					}
				}
			}
			if ($codonpos) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
					print(OUT "\tCharSet $temppartname$subpartno=$subpart{$temppartname . $subpartno}[0]-$subpart{$temppartname . $subpartno}[-1]\\3;\n");
				}
				$num ++;
			}
			else {
				$partno ++;
				push(@temppartname, $temppartname);
				print(OUT "\tCharSet $temppartname=$part2site{$temppartname}[0]-$part2site{$temppartname}[-1];\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . $partno . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) Correlation=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink TRatio=(" . join(',', @{$applyto[$i]}) . ") RevMat=(" . join(',', @{$applyto[$i]}) . ") StateFreq=(" . join(',', @{$applyto[$i]}) . ") Shape=(" . join(',', @{$applyto[$i]}) . ") PInvar=(" . join(',', @{$applyto[$i]}) . ") Correlation=(" . join(',', @{$applyto[$i]}) . ");\n");
		}
		print(OUT "\tPrSet ApplyTo=(all) RatePr=Variable;\n");
		print(OUT &makeMrBayesCommand($criterion, 1, @partname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# proportional model (codon position proportional)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_proportional_codonproportional.nex';
	}
	elsif ($partname eq 'whole' && @subpart > 0 && $subpartcriterion == 1 || $partname =~ /_P$/i && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_codonproportional.nex';
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		my @temppartname;
		if ($partname eq 'whole') {
			print(OUT &whole_gene_codonpos('MrBayes'));
			foreach my $temppartname (@partname) {
				if ($temppartname =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temppartname . $subpartno);
					}
				}
				else {
					push(@temppartname, $temppartname);
				}
			}
		}
		else {
			print(OUT &part_codonpos('MrBayes'));
			foreach my $subpartno (1,2,3) {
				push(@temppartname, $partname . $subpartno);
			}
		}
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) Correlation=(all);\n");
		print(OUT "\tPrSet ApplyTo=(all) RatePr=Variable;\n");
		print(OUT &makeMrBayesCommand($criterion, 1, @temppartname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# proportional model (codon position separate)
#	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && @subpart != 3 * @partname && $subpartcriterion == 1) {
#		$outfile = $partname . '_' . $criterion . '_proportional_codonseparate.nex';
#		open(OUT, "> MrBayes/$outfile");
#		print(OUT "#NEXUS\n\n");
#		print(OUT "Begin MrBayes;\n\tExecute ../data_$partname.nex;\n");
#		my $partno = 0;
#		my @applyto;
#		my @temppartname;
#		foreach my $temppartname (@partname) {
#			if ($temppartname =~ /_P$/i) {
#				foreach my $subpartno (1,2,3) {
#					$partno ++;
#					push(@temppartname, $temppartname . $subpartno);
#					print(OUT "\tCharSet $temppartname$subpartno=@{$subpart{$temppartname . $subpartno}};\n");
#				}
#			}
#			else {
#				$partno ++;
#				push(@applyto, $partno);
#				push(@temppartname, $temppartname);
#				print(OUT "\tCharSet $temppartname=@{$part2site{$temppartname}};\n");
#			}
#		}
#		print(OUT "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ';' . "\n");
#		print(OUT "\tSet Partition=Kakusan;\n");
#		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) BrLens=(all);\n");
#		print(OUT "\tLink BrLens=(" . join(',', @applyto) . ");\n");
#		print(OUT "\tPrSet ApplyTo=(" . join(',', @applyto) . ") RatePr=Variable;\n");
#		print(OUT &makeMrBayesCommand($criterion, @temppartname));
#		print(OUT "End;\n");
#		close(OUT);
#		undef($outfile);
#	}
	# separate model (codon position singlized)
	if ($partname eq 'whole' && @partname > 1 && $partcriterion == 1) {
		if (@subpart > 0) {
			$outfile = $partname . '_' . $criterion . '_separate_codonshared.nex';
		}
		else {
			$outfile = $partname . '_' . $criterion . '_separate.nex';
		}
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			my $codonpos;
			if ($outfile =~ /codonshared.nex$/) {
				my %modelMrBayes;
				foreach my $tempmodel (@modelformrbayes) {
					foreach my $tempheterogeneity (@heterogeneityformrbayes) {
						$modelMrBayes{$tempmodel . '_' . $tempheterogeneity} = 1;
					}
				}
				foreach my $tempmodel (sort({$criteria{$criterion}{$temppartname}{$a} <=> $criteria{$criterion}{$temppartname}{$b}} keys(%{$criteria{$criterion}{$temppartname}}))) {
					if ($modelMrBayes{$tempmodel}) {
						if ($tempmodel =~ /CodonPos/) {
							$codonpos = 1;
						}
						last;
					}
				}
			}
			if ($codonpos) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
					print(OUT "\tCharSet $temppartname$subpartno=$subpart{$temppartname . $subpartno}[0]-$subpart{$temppartname . $subpartno}[-1]\\3;\n");
				}
				$num ++;
			}
			else {
				$partno ++;
				push(@temppartname, $temppartname);
				print(OUT "\tCharSet $temppartname=$part2site{$temppartname}[0]-$part2site{$temppartname}[-1];\n");
			}
		}
		print(OUT "\tPartition Kakusan=" . $partno . ':' . join(',', @temppartname) . ';' . "\n");
		print(OUT "\tSet Partition=Kakusan;\n");
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) Correlation=(all) BrLens=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink TRatio=(" . join(',', @{$applyto[$i]}) . ") RevMat=(" . join(',', @{$applyto[$i]}) . ") StateFreq=(" . join(',', @{$applyto[$i]}) . ") Shape=(" . join(',', @{$applyto[$i]}) . ") PInvar=(" . join(',', @{$applyto[$i]}) . ") Correlation=(" . join(',', @{$applyto[$i]}) . ") BrLens=(" . join(',', @{$applyto[$i]}) . ");\n");
		}
		print(OUT &makeMrBayesCommand($criterion, 1, @partname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# separate model (codon position proportional)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_separate_codonproportional.nex';
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		print(OUT &whole_gene_codonpos('MrBayes'));
		my $partno = 0;
		my $num = 0;
		my @applyto;
		my @temppartname;
		foreach my $temppartname (@partname) {
			if ($temppartname =~ /_P$/i) {
				foreach my $subpartno (1,2,3) {
					$partno ++;
					push(@{$applyto[$num]}, $partno);
					push(@temppartname, $temppartname . $subpartno);
				}
				$num ++;
			}
			else {
				$partno ++;
				push(@temppartname, $temppartname);
			}
		}
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) Correlation=(all) BrLens=(all);\n");
		for (my $i = 0; $i < @applyto; $i ++) {
			print(OUT "\tLink BrLens=(" . join(',', @{$applyto[$i]}) . ");\n");
			print(OUT "\tPrSet ApplyTo=(" . join(',', @{$applyto[$i]}) . ") RatePr=Variable;\n");
		}
		print(OUT &makeMrBayesCommand($criterion, 1, @temppartname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
	# separate model (codon position separate)
	if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_separate_codonseparate.nex';
	}
	elsif ($partname eq 'whole' && @subpart > 0 && $subpartcriterion == 1 || $partname =~ /_P$/i && $subpartcriterion == 1) {
		$outfile = $partname . '_' . $criterion . '_codonseparate.nex';
	}
	if ($outfile) {
		open(OUT, "> MrBayes/$outfile");
		print(OUT &dataNexus($partname));
		print(OUT "\nBegin MrBayes;\n");
		my @temppartname;
		if ($partname eq 'whole') {
			print(OUT &whole_gene_codonpos('MrBayes'));
			foreach my $temppartname (@partname) {
				if ($temppartname =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temppartname . $subpartno);
					}
				}
				else {
					push(@temppartname, $temppartname);
				}
			}
		}
		else {
			print(OUT &part_codonpos('MrBayes'));
			foreach my $subpartno (1,2,3) {
				push(@temppartname, $partname . $subpartno);
			}
		}
		print(OUT "\tUnlink TRatio=(all) RevMat=(all) StateFreq=(all) Shape=(all) PInvar=(all) Correlation=(all) BrLens=(all);\n");
		print(OUT &makeMrBayesCommand($criterion, 1, @temppartname));
		print(OUT "End;\n");
		close(OUT);
		undef($outfile);
	}
}

sub makeMrBayesCommand {
	my %modelMrBayes;
	foreach my $model (@modelformrbayes) {
		foreach my $heterogeneity (@heterogeneityformrbayes) {
			$modelMrBayes{$model . '_' . $heterogeneity} = 1;
		}
	}
	my $partno;
	my $criterion = shift(@_);
	my $repeat = shift(@_);
	my $cmd;
	if (@_ > 1) {
		$partno = 1;
	}
	foreach my $partname (@_) {
		$cmd .= '[' . $partname . ']' . "\n";
		my $numpart;
		my $notbest;
		foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
			if ($partname eq 'whole' && $model =~ /GeneCodonPos/) {
				$numpart = @partname - (@subpart / 3) + @subpart;
			}
			elsif ($partname eq 'whole' && $model =~ /Gene/) {
				$numpart = @partname;
			}
			elsif ($partname =~ /_P$/i && $model =~ /CodonPos/) {
				$numpart = 3;
			}
			else {
				$numpart = 1;
			}
			$cmd .= '[ ' . sprintf("%-26s ", $model);
			$cmd .= sprintf("% 12.5e", $criteria{$criterion}{$partname}{$model});
			$cmd .= ' ]';
			if ($modelMrBayes{$model}) {
				unless ($notbest) {
					$cmd .= '  ';
				}
				else {
					$cmd .= '[ ';
				}
				$cmd .= &makeMrBayesLSet($model, $partno, $numpart, $partname);
				if ($repeat) {
					$cmd .= "\n";
					last;
				}
				unless ($notbest) {
					$notbest = 1;
				}
				else {
					$cmd .=' ]';
				}
			}
			$cmd .= "\n";
		}
		if ($partno) {
			$partno += $numpart;
		}
	}
	return($cmd);
}

sub makeMrBayesLSet {
	my ($model, $heterogeneity) = split(/_/, $_[0]);
	my $partno = $_[1];
	my $numpart = $_[2];
	my $partname = $_[3];
	my $mblset;
	my $mbprset;
	my $unlinkshape;
	if ($model eq 'JC69') {
		$mblset = 'Nst=1';
		$mbprset = 'StateFreqPr=Fixed(Equal)';
	}
	elsif ($model eq 'F81') {
		$mblset = 'Nst=1';
		$mbprset = 'StateFreqPr=Dirichlet(1,1,1,1)';
	}
	elsif ($model eq 'K80') {
		$mblset = 'Nst=2';
		$mbprset = 'StateFreqPr=Fixed(Equal)';
	}
	elsif ($model eq 'HKY85') {
		$mblset = 'Nst=2';
		$mbprset = 'StateFreqPr=Dirichlet(1,1,1,1)';
	}
	elsif ($model eq 'SYM') {
		$mblset = 'Nst=6';
		$mbprset = 'StateFreqPr=Fixed(Equal)';
	}
	elsif ($model eq 'GTR') {
		$mblset = 'Nst=6';
		$mbprset = 'StateFreqPr=Dirichlet(1,1,1,1)';
	}
	if ($heterogeneity eq 'Homogeneous') {
		$mblset .= ' Rates=Equal';
	}
	elsif ($heterogeneity eq 'Gamma') {
		$mblset .= ' Rates=Gamma NGammaCat=' . $ngammacat;
	}
	elsif ($heterogeneity eq 'AGamma') {
		$mblset .= ' Rates=ADGamma NGammaCat=' . $ngammacat;
	}
	elsif ($heterogeneity eq 'Invariant') {
		$mblset .= ' Rates=PropInv';
	}
	elsif ($heterogeneity eq 'GammaInvariant') {
		$mblset .= ' Rates=InvGamma NGammaCat=' . $ngammacat;
	}
	elsif ($heterogeneity eq 'Gene' || $heterogeneity eq 'GeneCodonPos' || $heterogeneity eq 'CodonPos') {
		$mblset .= ' Rates=Equal';
		$mbprset .= ' RatePr=Variable';
	}
	elsif ($heterogeneity eq 'Gene1Gamma' || $heterogeneity eq 'GeneCodonPos1Gamma' || $heterogeneity eq 'CodonPos1Gamma') {
		$mblset .= ' Rates=Gamma NGammaCat=' . $ngammacat;
		$mbprset .= ' RatePr=Variable';
	}
	elsif ($heterogeneity eq 'GeneNGamma' || $heterogeneity eq 'GeneCodonPosNGamma' || $heterogeneity eq 'CodonPos3Gamma') {
		$mblset .= ' Rates=Gamma NGammaCat=' . $ngammacat;
		$mbprset .= ' RatePr=Variable';
		$unlinkshape = 1;
	}
	elsif ($heterogeneity eq 'Gene1AGamma' || $heterogeneity eq 'GeneCodonPos1AGamma' || $heterogeneity eq 'CodonPos1AGamma') {
		$mblset .= ' Rates=ADGamma NGammaCat=' . $ngammacat;
		$mbprset .= ' RatePr=Variable';
	}
	elsif ($heterogeneity eq 'GeneNAGamma' || $heterogeneity eq 'GeneCodonPosNAGamma' || $heterogeneity eq 'CodonPos3AGamma') {
		$mblset .= ' Rates=ADGamma NGammaCat=' . $ngammacat;
		$mbprset .= ' RatePr=Variable';
		$unlinkshape = 1;
	}
	my $applyto;
	if ($partno) {
		my @applyto;
		for (my $i = 0; $i < $numpart; $i ++) {
			push(@applyto, $partno + $i);
		}
		$applyto = join(',', @applyto);
	}
	else {
		$applyto = 'all';
	}
	my $cmd;
	if ($partname =~ /_P$/i && $heterogeneity =~ /^CodonPos/ && !$partno) {
		$cmd = &part_codonpos('MrBayes');
	}
	elsif ($partname eq 'whole' && $heterogeneity =~ /^GeneCodonPos/ && !$partno) {
		$cmd = &whole_gene_codonpos('MrBayes');
	}
	elsif ($partname eq 'whole' && $heterogeneity =~ /^Gene/ && !$partno) {
		$cmd = &whole_gene('MrBayes');
	}
	$cmd =~ s/\t//g;
	$cmd =~ s/\n/ /g;
	$cmd .= "LSet ApplyTo=($applyto) NucModel=4by4 $mblset; PrSet ApplyTo=($applyto) $mbprset;";
	if ($unlinkshape) {
		$cmd .= " Unlink Shape=($applyto) Correlation=($applyto);";
	}
	return($cmd);
}

sub dataNexus {
	my $partname = shift(@_);
	my $data;
	unless (open(IN, "< data_$partname.nex")) {
		&errorMessage(__LINE__, "Cannot open \"$maindir/data_$partname.nex\".");
	}
	while (<IN>) {
		$data .= $_;
	}
	close(IN);
	return($data);
}

sub whole_gene {
	my $mode= shift(@_);
	my $cmd;
	if ($mode eq 'PAUP') {
		foreach my $partname (@partname) {
			$cmd .= "\tCharSet $partname=$part2site{$partname}[0]-$part2site{$partname}[-1];\n";
		}
		$cmd .= "\tCharPartition Kakusan=";
		for (my $i = 0; $i < @partname; $i ++) {
			$cmd .= "$partname[$i]:$partname[$i]";
			if ($i + 1 == @partname) {
				$cmd .= ";\n";
				last;
			}
			else {
				$cmd .= ',';
			}
		}
	}
	elsif ($mode eq 'MrBayes') {
		foreach my $partname (@partname) {
			$cmd .= "\tCharSet $partname=$part2site{$partname}[0]-$part2site{$partname}[-1];\n";
		}
		$cmd .= "\tPartition Kakusan=" . scalar(@partname) . ':' . join(',', @partname) . ";\n";
		$cmd .= "\tSet Partition=Kakusan;\n";
	}
	elsif ($mode eq 'Treefinder') {
		my $figure = length(@partname - (@subpart / 3) + @subpart);
		my @partitions;
		my $num = 1;
		foreach my $partname (@partname) {
			for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
				my @num = split('', sprintf("%0*d", $figure, $num));
				for (my $j = 0; $j < @num; $j ++) {
					$partitions[$j] .= $num[$j];
				}
			}
			$num ++;
		}
		for (my $i = 0; $i < @partitions; $i ++) {
			$cmd .= '"part' . ($i + 1) . '" ' . $partitions[$i] . "\n";
		}
	}
	return($cmd);
}

sub whole_gene_codonpos {
	my $mode = shift(@_);
	my $cmd;
	if (@partname == 1 && @subpart == 3) {
		$cmd = &part_codonpos($mode);
	}
	else {
		my @temppartname;
		if ($mode eq 'PAUP') {
			foreach my $temp (@partname) {
				if ($temp =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temp . $subpartno);
						$cmd .= "\tCharSet $temp$subpartno=$subpart{$temp . $subpartno}[0]-$subpart{$temp . $subpartno}[-1]\\3;\n";
					}
				}
				else {
					push(@temppartname, $temp);
					$cmd .= "\tCharSet $temp=$part2site{$temp}[0]-$part2site{$temp}[-1];\n";
				}
			}
			$cmd .= "\tCharPartition Kakusan=";
			for (my $i = 0; $i < @temppartname; $i ++) {
				$cmd .= "$temppartname[$i]:$temppartname[$i]";
				if ($i + 1 == @temppartname) {
					$cmd .= ";\n";
					last;
				}
				else {
					$cmd .= ',';
				}
			}
		}
		elsif ($mode eq 'MrBayes') {
			foreach my $temp (@partname) {
				if ($temp =~ /_P$/i) {
					foreach my $subpartno (1,2,3) {
						push(@temppartname, $temp . $subpartno);
						$cmd .= "\tCharSet $temp$subpartno=$subpart{$temp . $subpartno}[0]-$subpart{$temp . $subpartno}[-1]\\3;\n";
					}
				}
				else {
					push(@temppartname, $temp);
					$cmd .= "\tCharSet $temp=$part2site{$temp}[0]-$part2site{$temp}[-1];\n";
				}
			}
			$cmd .= "\tPartition Kakusan=" . scalar(@temppartname) . ':' . join(',', @temppartname) . ";\n";
			$cmd .= "\tSet Partition=Kakusan;\n";
		}
		elsif ($mode eq 'Treefinder') {
			my $figure = length(@partname - (@subpart / 3) + @subpart);
			my @partitions;
			my $num = 1;
			foreach my $partname (@partname) {
				if ($partname =~ /_P$/i) {
					for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
						my @num = split('',sprintf("%0*d", $figure, $num + ($i % 3)));
						for (my $j = 0; $j < @num; $j ++) {
							$partitions[$j] .= $num[$j];
						}
					}
					$num += 3;
				}
				else {
					for (my $i = 0; $i < @{$part2site{$partname}}; $i ++) {
						my @num = split('',sprintf("%0*d", $figure, $num));
						for (my $j = 0; $j < @num; $j ++) {
							$partitions[$j] .= $num[$j];
						}
					}
					$num ++;
				}
			}
			for (my $i = 0; $i < @partitions; $i ++) {
				$cmd .= '"part' . ($i + 1) . '" ' . $partitions[$i] . "\n";
			}
		}
	}
	return($cmd);
}

sub part_codonpos {
	my $mode = shift(@_);
	my $cmd;
	if ($mode eq 'PAUP') {
		$cmd = "\tCharSet First=1-.\\3;\n\tCharSet Second=2-.\\3;\n\tCharSet Third=3-.\\3;\n\tCharPartition Kakusan=First:First,Second:Second,Third:Third;\n";
	}
	elsif ($mode eq 'MrBayes') {
		$cmd = "\tCharSet First=1-.\\3;\n\tCharSet Second=2-.\\3;\n\tCharSet Third=3-.\\3;\n\tPartition Kakusan=3:First,Second,Third;\n\tSet Partition=Kakusan;\n";
	}
	elsif ($mode eq 'Treefinder') {
		$cmd = '"codonpos" ';
		my $length = shift(@_);
		unless ($length) {
			$length = $nchar;
		}
		for (my $i = 0; $i < $length; $i ++) {
			$cmd .= ($i % 3) + 1;
		}
		$cmd .= "\n";
	}
	return($cmd);
}

sub outforTreefinder {
	unless (mkdir('Treefinder')) {
		&errorMessage(__LINE__, "Cannot make \"$maindir/Treefinder\" directory.");
	}
	my @temppartname;
	if (@partname > 1) {
		@temppartname = @partname;
	}
	foreach my $criterion (keys(%criteria)) {
		foreach my $partname ('whole', @temppartname) {
			if ($partname eq 'whole' || $criteria{$criterion}{$partname}{'JC69_Homogeneous'} || $partname =~ /_P$/i && $codonseparateonly && $criteria{$criterion}{$partname . 1}{'JC69_Homogeneous'} && $criteria{$criterion}{$partname . 2}{'JC69_Homogeneous'} && $criteria{$criterion}{$partname . 3}{'JC69_Homogeneous'}) {
				&makeTFFiles($criterion, $partname);
			}
		}
	}
}

sub makeTFFiles {
	my ($criterion, $partname) = @_;
	my $partcriterion = 1;
	my $subpartcriterion = 1;
	if ($partname eq 'whole') {
		foreach my $temppartname (@partname) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_Homogeneous'}) {
				$partcriterion = 0;
			}
		}
		foreach my $temppartname (@subpart) {
			unless ($criteria{$criterion}{$temppartname}{'JC69_Homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	}
	elsif ($partname =~ /_P$/i) {
		foreach my $subpartno (1,2,3) {
			unless ($criteria{$criterion}{$partname . $subpartno}{'JC69_Homogeneous'}) {
				$subpartcriterion = 0;
			}
		}
	}
	# single model
	{
		my $suffix;
		my $prefix;
		my $outfile;
		if ($partname ne 'whole' && $partname !~ /_P$/i || $criteria{$criterion}{$partname}{'JC69_Homogeneous'}) {
			if ($partname eq 'whole' && @partname == 1 && @subpart > 0 || $partname =~ /_P$/i) {
				$suffix = '_codonshared';
			}
			else {
				$suffix = '_single';
			}
			$prefix = $partname . '_' . $criterion . $suffix;
			$outfile = $prefix . '.model';
			if ($criterion eq 'AIC') {
				open(OUT, "> Treefinder/$partname$suffix.tf");
				print(OUT &dataTF($partname));
				close(OUT);
			}
			&makeTLFiles(0, "$partname$suffix.tf", $prefix);
			open(OUT, "> Treefinder/$outfile");
			print(OUT &makeTFModelExpression($criterion, $partname));
			close(OUT);
		}
	}
	# proportional model (codon position singlized)
	if ($partname eq 'whole' && @partname > 1 && $partcriterion == 1) {
		my $suffix;
		if (@subpart > 0) {
			$suffix = '_proportional_codonshared';
			
		}
		else {
			$suffix = '_proportional';
		}
		my $prefix = $partname . '_' . $criterion . $suffix;
		my $outfile = $prefix . '.model';
		if ($criterion eq 'AIC') {
			open(OUT, "> Treefinder/$partname$suffix.tf");
			print(OUT &whole_gene('Treefinder'));
			print(OUT &dataTF($partname));
			close(OUT);
		}
		&makeTLFiles(1, "$partname$suffix.tf", $prefix);
		open(OUT, "> Treefinder/$outfile");
		print(OUT &makeTFModelExpression($criterion, @partname));
		close(OUT);
	}
	# proportional model (codon position proportional)
	{
		my $suffix;
		my $prefix;
		my $outfile;
		if ($partname eq 'whole' && @partname > 1 && @subpart > 0 && $subpartcriterion == 1) {
			$suffix = '_proportional_codonproportional';
		}
		elsif ($partname eq 'whole' && @subpart > 0 && $subpartcriterion == 1 || $partname =~ /_P$/i && $subpartcriterion == 1) {
			$suffix = '_codonproportional';
		}
		if ($suffix) {
			$prefix = $partname . '_' . $criterion . $suffix;
			$outfile = $prefix . '.model';
			if ($criterion eq 'AIC') {
				open(OUT, "> Treefinder/$partname$suffix.tf");
				if ($partname eq 'whole') {
					print(OUT &whole_gene_codonpos('Treefinder'));
				}
				else {
					print(OUT &part_codonpos('Treefinder', scalar(@{$part2site{$partname}})));
				}
				print(OUT &dataTF($partname));
				close(OUT);
			}
			&makeTLFiles(1, "$partname$suffix.tf", $prefix);
			open(OUT, "> Treefinder/$outfile");
			my @temppartname;
			if ($partname eq 'whole') {
				foreach my $temppartname (@partname) {
					if ($temppartname =~ /_P$/i) {
						foreach my $subpartno (1,2,3) {
							push(@temppartname, $temppartname . $subpartno);
						}
					}
					else {
						push(@temppartname, $temppartname);
					}
				}
			}
			else {
				foreach my $subpartno (1,2,3) {
					push(@temppartname, $partname . $subpartno);
				}
			}
			print(OUT &makeTFModelExpression($criterion, @temppartname));
			close(OUT);
		}
	}
}

sub makeTFModelExpression {
	my %modelTF;
	foreach my $model (@modelfortf) {
		foreach my $heterogeneity (@heterogeneityfortf) {
			$modelTF{$model . '_' . $heterogeneity} = 1;
		}
	}
	my $criterion = shift(@_);
	my $numpart = scalar(@_);
	my $cmd;
	if ($numpart > 1) {
		$cmd .= "{\n";
	}
	my $partno = 0;
	foreach my $partname (@_) {
		$partno ++;
		$cmd .= '% ' . $partname . "\n";
		my $notbest;
		foreach my $model (sort({$criteria{$criterion}{$partname}{$a} <=> $criteria{$criterion}{$partname}{$b}} keys(%{$criteria{$criterion}{$partname}}))) {
			$cmd .= "\t% " . sprintf("%-26s ", $model);
			$cmd .= sprintf("% 12.5e\n", $criteria{$criterion}{$partname}{$model});
			if ($modelTF{$model}) {
				unless ($notbest) {
					$cmd .= "\t\t  ";
				}
				else {
					$cmd .= "\t\t% ";
				}
				$cmd .= &makeTFModel($model);
				unless ($notbest) {
					$notbest = 1;
				}
				$cmd .= "\n";
			}
		}
		if ($partno < $numpart) {
			$cmd .= "\t,\n";
		}
	}
	if ($numpart > 1) {
		$cmd .= "}\n";
	}
	return($cmd);	
}

sub makeTFModel {
	my $model = shift(@_);
	my $cmd;
	if ($model =~ /^JC69/) {
		$cmd .= 'HKY[{1,1,1,1,1,1},{1,1,1,1}]';
	}
	elsif ($model =~ /^F81/) {
		$cmd .= 'HKY[{1,1,1,1,1,1},';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^K80/) {
		$cmd .= 'HKY[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^HKY85/) {
		$cmd .= 'HKY[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^TN93ef/) {
		$cmd .= 'TN[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^TN93/) {
		$cmd .= 'TN[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^J1ef/) {
		$cmd .= 'J1[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^J1/) {
		$cmd .= 'J1[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^J2ef/) {
		$cmd .= 'J2[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^J2/) {
		$cmd .= 'J2[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^TIMef/) {
		$cmd .= 'J3[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^TIM/) {
		$cmd .= 'J3[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^TVMef/) {
		$cmd .= 'TVM[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^TVM/) {
		$cmd .= 'TVM[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	elsif ($model =~ /^SYM/) {
		$cmd .= 'GTR[Optimum,{1,1,1,1}]';
	}
	elsif ($model =~ /^GTR/) {
		$cmd .= 'GTR[Optimum,';
		if ($basefreq eq 'Empirical') {
			$cmd .= 'Empirical]';
		}
		elsif ($basefreq eq 'Estimate') {
			$cmd .= 'Optimum]';
		}
	}
	if ($model =~ /Gamma$/) {
		$cmd .= ':G[Optimum]:' . $ngammacat;
	}
	elsif ($model =~ /GammaInvariant$/) {
		$cmd .= ':GI[Optimum]:' . $ngammacat;
	}
	elsif ($model =~ /Invariant$/) {
		$cmd .= ':I[Optimum]';
	}
	return($cmd);
}

sub dataTF {
	my $partname = shift(@_);
	my $data;
	unless (open(IN, "< data_$partname.nex")) {
		&errorMessage(__LINE__, "Cannot open \"$maindir/data_$partname.nex\".");
	}
	my $line = 0;
	while (<IN>) {
		$line ++;
		if ($line < 7) {
			next;
		}
		elsif ($line >= 7 && $line < 7 + $ntax) {
			s/^(\S+)/"$1"/;
			$data .= $_;
		}
		else {
			last;
		}
	}
	close(IN);
	return($data);
}

sub makeTLFiles {
	my $multilocus = shift(@_);
	my $datafile = shift(@_);
	my $prefix = shift(@_);
	my $optimum;
	my $save;
	my $bootstrap;
	if ($multilocus) {
		$optimum = "\tPartitionRates->Optimum,\n";
		$save = "\n\tSave[\n\t\treport|1|PartitionRates,\n\t\t\"$prefix\_optimum.rates\"\n\t],";
		$bootstrap = "\tPartitionRates->Load[\n\t\t\"$prefix\_optimum.rates\"\n\t],\n";
	}
	open(TL, '> Treefinder/' . $prefix . '_singlesearch.tl');
	print(TL <<"_END");
report:=ReconstructPhylogeny[
	\"$datafile\",
	SubstitutionModel->Load[
		\"$prefix.model\"
	],
$optimum	WithEdgeSupport->False,
	SearchDepth->2,
	AcceptFlatness->True
],
Oprec[
	20,
	SaveReport[
		AsReport[
			report
		],
		\"$prefix\_treesearch.log\"
	],
	Save[
		report|1|SubstitutionModel,
		\"$prefix\_optimum.model\"
	],$save
	SaveTree[
		AsTree[
			report|1|Phylogeny
		],
		\"$prefix\_optimum.nwk\",
		Format->\"NEWICK\"
	]
]
_END
	close(TL);
	open(TL, '> Treefinder/' . $prefix . '_shotgunsearch.tl');
	my $nstarts = $ntax - 3;
	my $nsteps = int((($ntax - 3) / 2) + 0.5);
	if ($nsteps < 5) {
		$nsteps = 5;
	}
	print(TL <<"_END");
SaveTreeList[
	GenerateStartTrees[
		\"$prefix\_optimum.nwk\",
		NStarts->$nstarts,
		NSteps->$nsteps,
		NTries->10
	],
	\"$prefix\_starts.nwk\",
	Format->\"NEWICK\"
],
newreport:=ReconstructPhylogeny[
	\"$datafile\",
	SubstitutionModel->Load[
		\"$prefix.model\"
	],
$optimum	StartTrees->\"$prefix\_starts.nwk\",
	WithEdgeSupport->False,
	SearchDepth->2,
	AcceptFlatness->True
],
pastreport:=LoadReport[
	\"$prefix\_treesearch.log\"
],
Spell[\"$prefix\\n\"],
If[
	pastreport|1|Likelihood>=newreport|1|Likelihood,
	Spell[\"There is no improvement.\\n\"],
	If[
		Oprec[20,ToString[pastreport|1|Likelihood]]==Oprec[20,ToString[newreport|1|Likelihood]],
		Spell[\"There is no improvement.\\n\"],
		Spell[\"The likelihood was improved.\\n\"]
	]
],
report:=Sort[
	JoinReports[
		newreport,
		pastreport
	],
	Descending->True
],
Oprec[
	20,
	SaveReport[
		AsReport[
			report
		],
		\"$prefix\_treesearch.log\"
	],
	Save[
		report|1|SubstitutionModel,
		\"$prefix\_optimum.model\"
	],$save
	SaveTree[
		AsTree[
			report|1|Phylogeny
		],
		\"$prefix\_optimum.nwk\",
		Format->\"NEWICK\"
	]
]
_END
	close(TL);
	open(TL, '> Treefinder/' . $prefix . '_bootstrap.tl');
	print(TL <<"_END");
BootstrapAnalysis[
	\"$datafile\",
	SubstitutionModel->Load[
		\"$prefix\_optimum.model\"
	],
$bootstrap	StartTrees->\"$prefix\_optimum.nwk\",
	SearchDepth->2,
	NReplicates->100,
	ConsensusLevel->50,
	SampleFile->\"$prefix\_bootreps.tree\"
],
boottreelist:=LoadTreeList[
	\"$prefix\_bootreps.tree\"
],
SaveTreeList[
	boottreelist,
	\"$prefix\_bootreps.nwk\",
	Format->\"NEWICK\"
],
SaveTree[
	BuildConsensusTree[
		boottreelist,
		WithEdgeLengthIntervals->False
	],
	\"$prefix\_consensus.nwk\",
	Format->\"NEWICK\"
]
_END
	close(TL);
}

sub errorMessage {
	my $lineno = shift(@_);
	my $message = shift(@_);
	print("ERROR!: line $lineno\n$message\n");
	if ($interactive) {
		print("Please press enter to abort.\n");
		<STDIN>;
	}
	else {
		print("If you want to read help message, run this script with \"--help\" option.\n");
	}
	exit(1);
}

sub helpMessage {
	print <<"_END";
Usage
=====

perl kakusan.pl [-hicsdfpgmatonl] inputfile

Command line options
====================

-h, --help
  Display this message and exit.

-i, --interactive=Enable|Disable
  Specify whether you want to use interactive mode or not. (default:
Disable)

-c, --chisq=Enable|Disable
  Specify whether you want to sort the candidate models by chi-square
test or not. (default: Enable)

-s, --siglevel=DECIMAL
  Specify significant level of chi-square test. (default: 0.05)

-w, --calcwhole=Enable|Disable
  Specify whether you want to consider single models on whole
sequences or not. (default: Disable)

-d, --dividecodon=Enable|Disable
  Specify whether you want to select a model on each codon position of
protein coding region or not. (default: Disable)

--codonseparateonly=Enable|Disable
  Specify whether you want to disable codon position combined analysis
or not. (default: Disable)

-f, --basefreq=Empirical|Estimate
  Specify base frequencies optimization approach. (default: Estimate)

-p, --partition=PARTITIONS
  Specify data partitions. (default: none)

-g, --ngammacat=INTEGER
  Specify the number of gamma rate categories. (default: 8)

-m, --ngam=Enable|Disable
  Specify whether you want to consider different Gamma shape for
each partition and codon position or not. (default: Disable)

-r, --agamma=Enable|Disable
  Specify whether you want to consider autocorrelated discrete gamma
heterogeneity or not. (default: Disable)

-a, --algorithm=NJJC|NJLogDet|SARandom|SAClosest|SAFurthest
  Specify tree generating algorithm for parameter optimization.
(default: NJJC)

-t, --tree=Common|Different
  If you specify 'Common', the parameters are optimized at a common
tree on each partition. If you specify 'Different', the parameters are
optimized at different tree on each partition. (default: Common)

-o, --output=MrBayes|PAUP|Treefinder|All
  Specify output options. (default: All)

-n, --numthreads=INTEGER
  Specify the number of threads. (default: 1)

-l, --likelihood=BASEML|TF|PAUP
  Specify likelihood calculation software. (default: BASEML)

Examples of data partition specification
=========================================

part1:1-200,part2:201-500,part3:501-550
part1_P:1-300
  \"_P\" means \"protein coding region\".
_END
	exit;
}
